# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the Ui code."""

from PyQt4 import QtGui, QtCore
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso import main, NO_OP
from ubuntu_kylin_sso.qt import ERROR_STYLE, maybe_elide_text, TITLE_STYLE
from ubuntu_kylin_sso.tests import (
    APP_NAME,
    HELP_TEXT,
    PING_URL,
    POLICY_URL,
    TC_URL,
    WINDOW_ID,
)
from ubuntu_kylin_sso.utils import compat

KWARGS = dict(app_name=APP_NAME, help_text=HELP_TEXT, ping_url=PING_URL,
              policy_url=POLICY_URL, tc_url=TC_URL, window_id=WINDOW_ID)

# is ok to access private method/attrs in tests
# pylint: disable=W0212


def build_string_for_pixels(label, width):
    """Return a random string that will be as big as 'width' in pixels."""
    char = 'm'
    fm = QtGui.QFontMetrics(label.font())
    pixel_width = fm.width(char)
    chars = int(width / pixel_width)
    result = char * chars
    assert pixel_width * chars <= width
    return result


class FakedObject(object):
    """Fake an object, record every call."""

    next_result = None
    exposed_methods = []

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._called = {}
        for i in self.exposed_methods:
            setattr(self, i, self._record_call(i))

    def _record_call(self, func_name):
        """Store values when calling 'func_name'."""

        def inner(*args, **kwargs):
            """Fake 'func_name'."""
            self._called[func_name] = (args, kwargs)
            return self.next_result

        return inner


class FakedSSOLoginBackend(FakedObject):
    """A faked sso_login backend."""

    exposed_methods = [
        'connect_to_signal',
        'generate_captcha',
        'login',
        'login_and_ping',
        'register_user',
        'request_password_reset_token',
        'set_new_password',
        'validate_email',
        'validate_email_and_ping',
    ]


class FakedBackend(object):
    """A faked backend."""

    sso_login = FakedSSOLoginBackend()


class FakePageUiStyle(object):
    """Fake the page."""

    def __init__(self):
        self.ui = self
        self.properties = {}
        super(FakePageUiStyle, self).__init__()

    def wizard(self):
        """Use itself as a fake wizard, too."""
        return self

    def text(self):
        """Return text."""
        return self.properties.get('text', '')

    # setText, setEnabled are inherited
    # pylint: disable=C0103
    def setText(self, text):
        """Save text."""
        self.properties['text'] = text

    def setEnabled(self, value):
        """Fake setEnabled."""
        self.properties['enabled'] = value

    def enabled(self):
        """Fake enabled."""
        return self.properties.get('enabled', False)

    def isEnabled(self):
        """Fake isEnabled."""
        return self.properties.get('enabled', False)

    def setProperty(self, key, val):
        """Fake setProperty to restyle some widget."""
        self.properties[key] = val

    def property(self, key):
        """Fake property from widget style."""
        return self.properties.get(key, False)

    def setDefault(self, val):
        """Fake button setDefault."""
        self.properties['default'] = val

    def style(self):
        """Fake style."""
        return self

    def unpolish(self, *args):
        """Fake unpolish."""
        self.properties['unpolish'] = len(args) > 0

    def polish(self, *args):
        """Fake polish."""
        self.properties['polish'] = len(args) > 0

    def setVisible(self, val):
        """Fake setVisible from Qt."""
        self.properties['visible'] = val

    def isVisible(self):
        """Fake isVisible from Qt."""
        return self.properties['visible']


class FakeOverlay(object):

    """A fake delay overlay."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.show_counter = 0
        self.hide_counter = 0
        self.args = (args, kwargs)

    def show(self):
        """Fake show."""
        self.show_counter += 1

    def hide(self):
        """Fake hide."""
        self.hide_counter += 1

    def resize(self, *args):
        """Fake resize."""


class FakeSignal(object):

    """A fake PyQt signal."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.target = None

    def connect(self, target):
        """Fake connect."""
        self.target = target

    def disconnect(self, *args):
        """Fake disconnect."""
        self.target = None

    def emit(self, *args):
        """Fake emit."""
        if self.target:
            self.target(*args)


class FakeOverlaySignal(FakeSignal):

    """Fake Signal for show and hide overlay."""

    def __init__(self, target):
        super(FakeOverlaySignal, self).__init__()
        self.target = target

    def connect(self, target):
        """We ignore the target and just call the function from the init."""


class FakeWizardPage(object):

    """A fake wizard page."""

    def __init__(self, *args, **kwargs):
        self.has_back_button = True

    # pylint: disable=C0103
    def initializePage(self):
        """Fake initializePage."""


class FakeMainWindow(object):

    """A fake MainWindow."""

    currentIdChanged = FakeSignal()
    loginSuccess = FakeSignal()
    registrationIncomplete = FakeSignal()
    registrationSuccess = FakeSignal()
    userCancellation = FakeSignal()
    shown = False
    SYNC_NOW_OR_LATER_PAGE = 4
    CONGRATULATIONS_PAGE = 5
    folders_page_id = 6
    local_folders_page_id = 7

    def __init__(self, close_callback=None):
        self.button_texts = []
        self.button_layout = None
        self.options = []
        self.overlay = FakeOverlay()
        self.local_folders_page = FakeWizardPage()
        self.folders_page = FakeWizardPage()
        self.app_name = APP_NAME

    def show(self):
        """Fake method."""
        self.shown = True

    # pylint: disable=C0103
    def setButtonText(self, button, text):
        """Fake setButtonText."""
        self.button_texts.append((button, text))

    def setButtonLayout(self, layout):
        """Fake setButtonText."""
        self.button_layout = layout

    def button(self, button):
        """Fake button method to obtain the wizard buttons."""
        return QtGui.QPushButton()

    def setOption(self, *args):
        """Fake setOption."""
        self.options.append(args)

    def next(self):
        """Fake next."""

    def reject(self):
        """Fake reject."""


class FakeController(object):

    """A fake controller for the tests."""

    is_fake = True
    set_next_validation = lambda *a, **kw: None

    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)

    # pylint: disable=C0103
    def setupUi(self, view):
        """Fake the setup."""
    # pylint: enable=C0103


class FakeWizard(object):

    """Replace wizard() function on wizard pages."""

    customButtonClicked = QtCore.QObject()
    sign_in_page_id = 1
    setup_account_id = 2
    current_user_id = 3
    email_verification_id = 4
    success_id = 5
    error_id = 6
    forgotten_id = 7
    reset_password_id = 8

    def __init__(self):
        self.overlay = FakeOverlay()
        self.called = []
        self.buttons = {}
        self.buttons_text = {}
        self._next_id = -1
        self.sign_in_page = object()
        self.setup_account = object()
        self.current_user = object()
        self.email_verification = object()
        self.success = object()
        self.error = object()
        self.forgotten = object()
        self.reset_password = object()
        self._pages = {
            self.sign_in_page: self.sign_in_page_id,
            self.setup_account: self.setup_account_id,
            self.current_user: self.current_user_id,
            self.email_verification: self.email_verification_id,
            self.success: self.success_id,
            self.error: self.error_id,
            self.forgotten: self.forgotten_id,
            self.reset_password: self.reset_password_id,
        }

    # Invalid name "setButtonLayout", "setOption"
    # pylint: disable=C0103

    def setButtonText(self, key, value):
        """Fake setButtonText."""
        self.buttons_text[key] = value

    def setButtonLayout(self, *args, **kwargs):
        """Fake the functionality of setButtonLayout on QWizard class."""
        self.called.append(('setButtonLayout', (args, kwargs)))

    def setOption(self, *args, **kwargs):
        """Fake the functionality of setOption on QWizard class."""
        self.called.append(('setOption', (args, kwargs)))

    # pylint: enable=C0103

    def button(self, button_id):
        """Fake the functionality of button on QWizard class."""
        return self.buttons.setdefault(button_id, QtGui.QPushButton())

    def next(self):
        """Fake next."""
        self.called.append('next')

    def page(self, p_id):
        """Fake page method."""
        return self

    def set_titles(self, title):
        """Fake set_titles."""
        self.called.append(('set_titles', title))


class FakeWizardButtonStyle(FakeWizard):

    """Fake Wizard with button style implementation."""

    SIGN_IN_PAGE_ID = 1

    # pylint: disable=C0103
    def __init__(self):
        super(FakeWizardButtonStyle, self).__init__()
        self.data = {}
        self.customButtonClicked = self

    def setDefault(self, value):
        """Fake setDefault for button."""
        self.data['default'] = value

    def isDefault(self):
        """Fake isDefault."""
        return self.data['default']
    # pylint: enable=C0103

    def connect(self, func):
        """Fake customButtonClicked connect."""
        self.data['connect'] = func

    def disconnect(self, func):
        """Fake customButtonClicked disconnect."""
        self.data['disconnect'] = func

    def button(self, button_id):
        """Fake the functionality of button on QWizard class."""
        return self

    def style(self):
        """Fake button style."""
        return self

    def next(self):
        """Fake next for wizard."""
        self.data['next'] = True


class BaseTestCase(TestCase):
    """The base test case."""

    kwargs = None
    ui_class = None
    ui_signals = ()
    ui_backend_signals = ()
    ui_wizard_class = FakeWizard

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        self._called = False
        backend = FakedBackend()
        self.sso_login_backend = backend.sso_login
        self.patch(main, 'get_sso_client',
                   lambda *a: defer.succeed(backend))

        self.app_name = APP_NAME
        self.ping_url = PING_URL
        self.signal_results = []

        # self.ui_class is not callable
        # pylint: disable=E1102, C0103, W0212
        self.ui = None
        self.wizard = None
        if self.ui_class is not None:
            kwargs = dict(KWARGS) if self.kwargs is None else self.kwargs
            self.ui = self.ui_class(**kwargs)

            for signal in self.ui_signals:
                self.patch(self.ui_class, signal, FakeSignal())

            if self.ui_wizard_class is not None:
                self.wizard = self.ui_wizard_class()
                self.patch(self.ui, 'wizard', lambda: self.wizard)

            self.ui.show()
            self.addCleanup(self.ui.hide)

    def _set_called(self, *args, **kwargs):
        """Store 'args' and 'kwargs' for test assertions."""
        self._called = (args, kwargs)

    def assert_backend_called(self, method, *args, **kwargs):
        """Check that 'method(*args, **kwargs)' was called in the backend."""
        self.assertIn(method, self.ui.backend._called)

        call = self.ui.backend._called[method]
        self.assertEqual(call[0], args)

        reply_handler = call[1].pop('reply_handler')
        self.assertEqual(reply_handler, NO_OP)

        error_handler = call[1].pop('error_handler')
        self.assertEqual(error_handler.func, self.ui._handle_error)

        self.assertEqual(call[1], kwargs)

    def assert_signal_emitted(self, signal, signal_args,
                              trigger, *args, **kwargs):
        """Check that 'trigger(*signal_args)' emits 'signal(*signal_args)'."""
        signal.connect(lambda *a: self.signal_results.append(a))

        trigger(*args, **kwargs)

        self.assertEqual(self.signal_results, [signal_args])

    def assert_title_correct(self, title_label, expected, max_width):
        """Check that the label's text is equal to 'expected'."""
        label = QtGui.QLabel()
        maybe_elide_text(label, expected, max_width)

        self.assertEqual(TITLE_STYLE % compat.text_type(label.text()),
                         compat.text_type(title_label.text()))
        self.assertEqual(compat.text_type(label.toolTip()),
                         compat.text_type(title_label.toolTip()))
        self.assertTrue(title_label.isVisible())

    def assert_subtitle_correct(self, subtitle_label, expected, max_width):
        """Check that the subtitle is equal to 'expected'."""
        label = QtGui.QLabel()
        maybe_elide_text(label, expected, max_width)

        self.assertEqual(compat.text_type(label.text()),
                         compat.text_type(subtitle_label.text()))
        self.assertEqual(compat.text_type(label.toolTip()),
                         compat.text_type(subtitle_label.toolTip()))
        self.assertTrue(subtitle_label.isVisible())

    def assert_error_correct(self, error_label, expected, max_width):
        """Check that the error 'error_label' displays 'expected' as text."""
        label = QtGui.QLabel()
        maybe_elide_text(label, expected, max_width)

        self.assertEqual(ERROR_STYLE % compat.text_type(label.text()),
                         compat.text_type(error_label.text()))
        self.assertEqual(compat.text_type(label.toolTip()),
                         compat.text_type(error_label.toolTip()))
        self.assertTrue(error_label.isVisible())

    def get_pixmap_data(self, pixmap):
        """Get the raw data of a QPixmap."""
        byte_array = QtCore.QByteArray()
        array_buffer = QtCore.QBuffer(byte_array)
        pixmap.save(array_buffer, "PNG")
        return byte_array

    # Invalid name "assertEqualPixmap"
    # pylint: disable=C0103

    def assertEqualPixmaps(self, pixmap1, pixmap2):
        """Compare two Qt pixmaps."""
        d1 = self.get_pixmap_data(pixmap1)
        d2 = self.get_pixmap_data(pixmap2)
        self.assertEqual(d1, d2)

    # pylint: enable=C0103

    @defer.inlineCallbacks
    def test_setup_page(self):
        """Test the backend signal connection."""
        if self.ui is None or getattr(self.ui, 'setup_page', None) is None:
            return

        called = []
        self.patch(self.ui, '_setup_signals',
                   lambda *a: called.append('_setup_signals'))
        self.patch(self.ui, '_connect_ui',
                   lambda *a: called.append('_connect_ui'))
        self.patch(self.ui, '_set_translated_strings',
                   lambda *a: called.append('_set_translated_strings'))

        yield self.ui.setup_page()

        self.assertEqual(self.ui.backend, self.sso_login_backend)

        for signal in self.ui_backend_signals:
            self.assertIn(signal, self.ui._signals)
            self.assertTrue(callable(self.ui._signals[signal]))

        expected = ['_set_translated_strings', '_connect_ui', '_setup_signals']
        self.assertEqual(expected, called)


class PageBaseTestCase(BaseTestCase):

    """BaseTestCase with some specialization for the Wizard Pages."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(PageBaseTestCase, self).setUp()
        self._overlay_show_counter = 0
        self._overlay_hide_counter = 0

        if self.ui is not None:
            fake_show_overlay_signal = FakeOverlaySignal(
                self._show_overlay_slot)
            fake_hide_overlay_signal = FakeOverlaySignal(
                self._hide_overlay_slot)
            self.patch(self.ui, "processingStarted", fake_show_overlay_signal)
            self.patch(self.ui, "processingFinished", fake_hide_overlay_signal)

    def _show_overlay_slot(self):
        """Fake show overlay slot."""
        self._overlay_show_counter += 1

    def _hide_overlay_slot(self):
        """Fake hide overlay slot."""
        self._overlay_hide_counter += 1

    def assert_signal_emitted(self, signal, signal_args,
                              trigger, *args, **kwargs):
        """Check that 'trigger(*args, **kwargs)' emits 'signal(*signal_args)'.

        Also check that the _overlay_hide_counter was increased by one, and
        that the ui is enabled.

        """
        super(PageBaseTestCase, self).assert_signal_emitted(signal,
            signal_args, trigger, *args, **kwargs)
        self.assertEqual(self._overlay_hide_counter, 1)
        self.assertTrue(self.ui.isEnabled())

    # pylint: disable=W0221

    def assert_title_correct(self, expected):
        """Check that the title is equal to 'expected'."""
        check = super(PageBaseTestCase, self).assert_title_correct
        check(self.ui.header.title_label, expected,
              self.ui.header.max_title_width)

    def assert_subtitle_correct(self, expected):
        """Check that the subtitle is equal to 'expected'."""
        check = super(PageBaseTestCase, self).assert_subtitle_correct
        check(self.ui.header.subtitle_label, expected,
              self.ui.header.max_subtitle_width)

    # pylint: enable=W0221



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Utils to be used by the UI modules."""

import argparse
import os
import re

from ubuntu_kylin_sso.logger import setup_logging

from ubuntu_kylin_sso.utils.translation import get_gettext

source_path = os.path.join(os.path.dirname(__file__),
                           os.path.pardir, os.path.pardir,
                           'build', 'mo')
# _ = get_gettext('ubuntu-sso-client',
_ = get_gettext('ubuntu-kylin-sso-client',
                fallback_path=os.path.abspath(source_path))

logger = setup_logging('ubuntu_kylin_sso.utils.ui')


# Undefined variable '_', pylint: disable=E0602

# all the text that is used in the gui
AGREE_TO_PRIVACY_POLICY = _('By signing up to {app_name} you agree to '
    'our {privacy_policy}')
AGREE_TO_TERMS = _('By signing up to {app_name} you agree to '
    'our {terms_and_conditions}')
AGREE_TO_TERMS_AND_PRIVACY_POLICY = AGREE_TO_TERMS + _(' and {privacy_policy}')
CANCEL_BUTTON = _('Cancel')
CAPTCHA_SOLUTION_ENTRY = _('Type the characters above')
CAPTCHA_LOAD_ERROR = _('There was a problem getting the captcha, '
                       'reloading...')
CAPTCHA_RELOAD_MESSAGE = _('If you can\'t read this then %(reload_link)s '
    'this page')
CAPTCHA_RELOAD_TEXT = _('refresh')
CAPTCHA_RELOAD_TOOLTIP = _('Reload')
CAPTCHA_REQUIRED_ERROR = _('The captcha is a required field')
CLOSE_AND_SETUP_LATER = _('Close window and set up later')
CONGRATULATIONS = _("Congratulations, {app_name} is installed!")
CONNECT_HELP_LABEL = _('To connect this computer to %(app_name)s enter your '
    'details below.')
CREATE_ACCOUNT_LABEL = _('Register with {app_name}.')
EMAIL_LABEL = EMAIL1_ENTRY = _('Email address')
EMAIL_USERNAME = _('Email address or username')
EMAIL2_ENTRY = _('Re-type Email address')
EMAIL_INVALID = _('The email must be a valid email address.')
EMAIL_MISMATCH = _('The email addresses don\'t match, please double check '
                   'and try entering them again.')
EMAIL = _("Email")
EMAIL_MATCH = _("The email addresses do not match")
EMAIL_TOKEN_ENTRY = _('Enter code verification here')
EMPTY_NAME = _("Please enter your name")
ERROR = _('The process did not finish successfully.')
ERROR_EMAIL_TOKEN = 'email_token'
EXISTING_ACCOUNT_CHOICE_BUTTON = _('Sign me in with my existing account')
FIELD_REQUIRED = _('This field is required.')
FORGOTTEN_PASSWORD_BUTTON = _('I\'ve forgotten my password')
FORGOTTEN_PASSWORD_TITLE = _('Reset password')
FORGOTTEN_PASSWORD_SUBTITLE = _('To reset your {app_name} password, enter '
    'your registered email address below. We will send instructions to reset '
    'your password.')
INVALID_EMAIL = _("Please enter a valid email address")
GENERIC_BACKEND_ERROR = _('There was a problem accessing the Ubuntu'
                          ' One backend.')
JOIN_HEADER_LABEL = _('Create %(app_name)s account')
LOADING = _('Loading...')
LOADING_OVERLAY = _('Getting information, please wait...')
LOGIN_BUTTON_LABEL = _('Already have an account? Click here to sign in')
LOGIN_EMAIL_ENTRY = _('Email address')
LOGIN_HEADER_LABEL = _('Connect to %(app_name)s')
LOGIN_PASSWORD_ENTRY = _('Password')
LOGIN_PASSWORD_LABEL = LOGIN_PASSWORD_ENTRY
LOGIN_TITLE = _('Sign In to {app_name}')
LOGIN_SUBTITLE = CONNECT_HELP_LABEL
NAME_ENTRY = _('Name')
NAME_INVALID = _('The name must not be empty.')
NEXT = _('Next')
FINISH = _('Finish')
NETWORK_DETECTION_TITLE = _('Network detection')
NETWORK_DETECTION_WARNING = _('Are you online? We can\'t detect an internet '
    'connection - you will need to be connected to set up %(app_name)s')
ONE_MOMENT_PLEASE = _('One moment please...')
PASSWORD = _("Create a password")
PASSWORD_CHANGED = _('Your password was successfully changed.')
PASSWORD_DIGIT = _("At least one number")
PASSWORD1_ENTRY = RESET_PASSWORD1_ENTRY = _('Password')
PASSWORD2_ENTRY = RESET_PASSWORD2_ENTRY = _('Re-type Password')
# weaken the requirement of password strength. Robert
# PASSWORD_HELP = _('The password must have a minimum of 8 characters and '
#     'include one uppercase character and one number.')
# PASSWORD_LENGTH = _("At least 8 characters")
PASSWORD_UPPER = _("At least one uppercase letter")
PASSWORD_HELP = _('The password must have a minimum of 8 characters.')
PASSWORD_LENGTH = _("At least 8 characters")
PASSWORD_MATCH = _("Passwords don't match")
PASSWORD_MISMATCH = _('The passwords don\'t match, please double check '
    'and try entering them again.')
PASSWORD_MUST_CONTAIN = _("Your password must contain")
PASSWORD_TOO_WEAK = _('The password is too weak.')
PRIVACY_POLICY_TEXT = _("Privacy Policy")
PROXY_CREDS_CONNECTION = _('Connecting to:')
PROXY_CREDS_DIALOG_TITLE = _('Proxy Settings')
PROXY_CREDS_ERROR = _('Incorrect login details. Please try again.')
PROXY_CREDS_EXPLANATION = _('Please provide login details.')
PROXY_CREDS_HEADER = _('You are connecting through a proxy.')
PROXY_CREDS_HELP_BUTTON = _('Get Help With Proxies')
PROXY_CREDS_PSWD_LABEL = _('Proxy password:')
PROXY_CREDS_SAVE_BUTTON = _('Save and Connect')
PROXY_CREDS_USER_LABEL = _('Proxy username:')
RESET_TITLE = _("Reset password")
RESET_SUBTITLE = _('A password reset code has been sent to your e-mail. '
    'Please enter the code below along with your new password.')
RETYPE_EMAIL = _("Retype email")
RETYPE_PASSWORD = _("Retype password")
REQUEST_PASSWORD_TOKEN_LABEL = _('To reset your %(app_name)s password,'
                                 ' enter your email address below:')
REQUEST_PASSWORD_TOKEN_TECH_ERROR = _('We are very Sorry! The service that'
    ' signs you on is not responding right now\nPlease try again or'
    ' come back in a few minutes.')
REQUEST_PASSWORD_TOKEN_WRONG_EMAIL = _('Sorry we did not recognize the email'
    ' address.')
RESET_CODE_ENTRY = _('Reset code')
RESET_EMAIL_ENTRY = _('Email address')
RESET_PASSWORD = RESET_TITLE
SET_NEW_PASSWORD_LABEL = _('A password reset code has been sent to '
    '%(email)s.\nPlease enter the code below along with your new password.')
SET_UP_ACCOUNT_BUTTON = _('Set Up Account')
SET_UP_ACCOUNT_CHOICE_BUTTON = _('I don\'t have an account yet - sign me up')
SIGN_IN_BUTTON = _('Sign In')
SIGN_IN_LABEL = _('Log-in with my existing account.')
SSL_APPNAME_HELP = _('the appname whose ssl error we are going to show.')
SSL_CERT_DETAILS = _('Certificate details')
SSL_CONNECT_BUTTON = _('Connect')
SSL_DETAILS_HELP = _('the details ssl certificate we are going to show.')
SSL_DETAILS_TEMPLATE = ('Organization:\t%(organization)s\n'
                        'Common Name:\t%(common_name)s\n'
                        'Locality Name:\t%(locality_name)s\n'
                        'Unit:\t%(unit)s\n'
                        'Country:\t%(country_name)s\n'
                        'State or Province:\t%(state_name)s')
SSL_DESCRIPTION = _('Open the SSL certificate UI.')
SSL_DIALOG_TITLE = _('SSL Certificate Not Valid')
SSL_DOMAIN_HELP = _('the domain whose ssl certificate we are going to show.')
SSL_EXPLANATION = _('You are trying to connect to a proxy server on'
    ' %(domain)s. This server uses a secure connection, and the SSL '
    'certificate is not valid because:')
SSL_FIRST_REASON = _('The certificate has not been verified')
SSL_HEADER = _('Do you want to connect to this server?')
SSL_HELP_BUTTON = _('Get Help With SSL')
SSL_NOT_SURE = _('If you are not sure about this server, do not use it to'
                 ' connect to %(app_name)s.')
SSL_REMEMBER_DECISION = _('Remember my settings for this certificate.')
SSL_SECOND_REASON = _('The name on the certificate isn\'t valid or doesn\'t'
    ' match the name of the site')
SSL_THIRD_REASON = _('The certificate has expired')
SUCCESS = _('You are now logged into %(app_name)s.')
SURNAME_ENTRY = _('Surname')
TERMS_TEXT = _("Terms of Service")
TITLE = REGISTER_TITLE = _("Sign Up to {app_name}")
TC_BUTTON = _('Show Terms & Conditions')
TC_NOT_ACCEPTED = _('Agreeing to the %(app_name)s Terms & Conditions is '
    'required to subscribe.')
TOS_LABEL = _("You can also find these terms at <a href='%(url)s'>%(url)s</a>")
TRY_AGAIN_BUTTON = _('Try again')
UNKNOWN_ERROR = _('There was an error when trying to complete the '
    'process. Please check the information and try again.')
VERIFICATION_CODE = _('Verification code')
# VERIFY_EMAIL_CONTENT = _('Check %(email)s for an email from Ubuntu One. '
VERIFY_EMAIL_CONTENT = _('Check %(email)s for an email from Ubuntu Kylin. '
    'This message contains a verification code. Enter the code in '
    'the field below and click OK to complete creating your %(app_name)s '
    'account.')
VERIFY_EMAIL_TITLE = _('Enter verification code')
VERIFY_EMAIL_LABEL = ('<b>%s</b>\n\n' % VERIFY_EMAIL_TITLE +
                      VERIFY_EMAIL_CONTENT)
YES_TO_TC = _('I agree with the %(app_name)s terms and conditions')
YES_TO_UPDATES = _('Yes! Email me %(app_name)s tips and updates.')

WINDOW_TITLE_COMMON = _('Register & Sign In')
# pylint: enable=E0602


def get_password_strength(password):
    """Return the strength of the password.

    This function returns the strength of the password so that ui elements
    can show the user how good his password is. The logic used is the
    following:

    * 1 extra point for 4 chars passwords
    * 1 extra point for 8 chars passwords
    * 1 extra point for more than 11 chars passwords.
    * 1 extra point for passwords with at least one number.
    * 1 extra point for passwords for lower and capital chars.
    * 1 extra point for passwords with a special char.

    A passwords starts with 0 and the extra points are added accordingly.
    """
    score = 0
    if len(password) < 1:
        return 0
    if len(password) < 4:
        score = 1
    if len(password) >= 8:
        score += 1
    if len(password) >= 11:
        score += 1
    if re.search('\d+', password):
        score += 1
    if re.search('[a-z]', password) and re.search('[A-Z]', password):
        score += 1
    if re.search('.[!,@,#,$,%,^,&,*,?,_,~,-,£,(,)]', password):
        score += 1
    return score


def is_min_required_password(password):
    """Return if the password meets the minimum requirements."""
    # if len(password) < 8 or \
    #         re.search('[A-Z]', password) is None or \
    #         re.search('\d+', password) is None:
    #     return False
    # return True

    # weaken the password requirement. Robert
    if len(password) < 8:
        return False
    return True


def is_correct_email(email_address):
    """Return if the email is correct."""
    return '@' in email_address


def parse_args():
    """Parse sys.argv options."""
    parser = argparse.ArgumentParser(description='Open the Ubuntu One UI.')
    parser.add_argument('--app_name', default='',
        help='the name of the application to retrieve credentials for')
    parser.add_argument('--ping_url', default='',
        help='a link to be used as the ping url (to notify about new tokens)')
    parser.add_argument('--policy_url', default='',
        help='a link to be used as Privacy Policy url')
    parser.add_argument('--tc_url', default='',
        help='a link to be used as Terms & Conditions url')
    parser.add_argument('--help_text', default='',
        help='extra text that will be shown below the headers')
    parser.add_argument('--window_id', type=int, default=0,
        help='the window id to be set transient for the Ubuntu One dialogs')
    parser.add_argument('--login_only', action='store_true', default=False,
        help='whether the Ubuntu One UI should only offer login or not')

    args = parser.parse_args()
    return args



# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""A webclient backend that uses QtNetwork."""

from __future__ import unicode_literals

import glob
import os
import sys
from io import StringIO

# pylint: disable=E0611
from PyQt4.QtCore import (
    QBuffer,
    QCoreApplication,
    QUrl,
)
# pylint: enable=E0611
from PyQt4.QtNetwork import (
    QNetworkAccessManager,
    QNetworkProxy,
    QNetworkProxyFactory,
    QNetworkReply,
    QNetworkRequest,
    QSslCertificate,
    QSslConfiguration,
    QSslSocket,
    QSsl,
)
from twisted.internet import defer

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import get_cert_dir
from ubuntu_kylin_sso.utils.webclient.common import (
    BaseWebClient,
    HeaderDict,
    ProxyUnauthorizedError,
    Response,
    UnauthorizedError,
    WebClientError,
)
from ubuntu_kylin_sso.utils.webclient import gsettings

logger = setup_logging("ubuntu_kylin_sso.utils.webclient.qtnetwork")


def build_proxy(settings_groups):
    """Create a QNetworkProxy from these settings."""
    proxy_groups = [
        ("socks", QNetworkProxy.Socks5Proxy),
        ("https", QNetworkProxy.HttpProxy),
        ("http", QNetworkProxy.HttpProxy),
    ]
    for group, proxy_type in proxy_groups:
        if group not in settings_groups:
            continue
        settings = settings_groups[group]
        if "host" in settings and "port" in settings:
            return QNetworkProxy(proxy_type,
                                 hostName=settings.get("host", ""),
                                 port=settings.get("port", 0),
                                 user=settings.get("username", ""),
                                 password=settings.get("password", ""))
    logger.error("No proxy correctly configured.")
    return QNetworkProxy(QNetworkProxy.DefaultProxy)


class WebClient(BaseWebClient):
    """A webclient with a qtnetwork backend."""

    proxy_instance = None

    def __init__(self, *args, **kwargs):
        """Initialize this instance."""
        super(WebClient, self).__init__(*args, **kwargs)
        self.nam = QNetworkAccessManager(QCoreApplication.instance())
        self.nam.finished.connect(self._handle_finished)
        self.nam.authenticationRequired.connect(self._handle_authentication)
        self.nam.proxyAuthenticationRequired.connect(self.handle_proxy_auth)
        self.nam.sslErrors.connect(self._handle_ssl_errors)
        self.replies = {}
        self.proxy_retry = False
        self.setup_proxy()

        # Force Qt to load the system certificates
        QSslSocket.setDefaultCaCertificates(QSslSocket.systemCaCertificates())
        # Apply our local certificates as the SSL configuration to be used
        # for all QNetworkRequest calls.
        self.ssl_config = QSslConfiguration.defaultConfiguration()
        ca_certs = self.ssl_config.caCertificates()
        try:
            for path in glob.glob(os.path.join(get_cert_dir(),
                                                #"UbuntuOne*.pem")):
                                                "UbuntuKylin_CA.pem")):
                #print path
                with open(path) as f:
                    cert = QSslCertificate(f.read())
                    if cert.isValid():
                        ca_certs.append(cert)
                        # if path.find('ukid_ca'):
                        #     print 'ukid_ca'
                        #     QSslSocket.addCaCertificate(cert)
                        #print "cert.isValid()"
                        #print self._get_certificate_details(cert)
                        #print self._get_certificate_host(cert)
                    else:
                        #print "invalid certificate: {}".format(path)
                        logger.error("invalid certificate: {}".format(path))
        except (IndexError, IOError) as err:
            #print 'Unable to configure SSL certificates: {}'
            raise WebClientError(
                    "Unable to configure SSL certificates: {}".format(err))

        #self.ssl_config.setProtocol(QSsl.TlsV1) # TlsV1_0 TlsV1_2
        #print 'BEFORE self.ssl_config.setCaCertificates(ca_certs)'
        #print self._get_certificate_details(ca_certs.count)
        self.ssl_config.setCaCertificates(ca_certs)

    def _set_proxy(self, proxy):
        """Set the proxy to be used."""
        QNetworkProxy.setApplicationProxy(proxy)
        self.nam.setProxy(proxy)

    def setup_proxy(self):
        """Setup the proxy settings if needed."""
        # QtNetwork knows how to use the system settings on both Win and Mac
        if sys.platform.startswith("linux"):
            settings = gsettings.get_proxy_settings()
            enabled = len(settings) > 0
            if enabled and WebClient.proxy_instance is None:
                proxy = build_proxy(settings)
                self._set_proxy(proxy)
                WebClient.proxy_instance = proxy
            elif enabled and WebClient.proxy_instance:
                logger.info("Proxy already in use.")
            else:
                logger.info("Proxy is disabled.")
        else:
            if WebClient.proxy_instance is None:
                logger.info("Querying OS for proxy.")
                QNetworkProxyFactory.setUseSystemConfiguration(True)

    def handle_proxy_auth(self, proxy, authenticator):
        """Proxy authentication is required."""
        logger.info("auth_required %r, %r", self.proxy_username,
                                            proxy.hostName())
        if (self.proxy_username is not None and
                self.proxy_username != str(authenticator.user())):
            authenticator.setUser(self.proxy_username)
            WebClient.proxy_instance.setUser(self.proxy_username)
        if (self.proxy_password is not None and
                self.proxy_password != str(authenticator.password())):
            authenticator.setPassword(self.proxy_password)
            WebClient.proxy_instance.setPassword(self.proxy_password)

    def _perform_request(self, request, method, post_buffer):
        """Return a deferred that will be fired with a Response object."""
        d = defer.Deferred()
        if method == "GET":
            reply = self.nam.get(request)
        elif method == "HEAD":
            reply = self.nam.head(request)
        else:
            reply = self.nam.sendCustomRequest(request, method, post_buffer)
        self.replies[reply] = d
        return d

    @defer.inlineCallbacks
    def request(self, iri, method="GET", extra_headers=None,
                oauth_credentials=None, post_content=None):
        """Return a deferred that will be fired with a Response object."""
        uri = self.iri_to_uri(iri)
        request = QNetworkRequest(QUrl(uri))
        request.setSslConfiguration(self.ssl_config)
        headers = yield self.build_request_headers(uri, method, extra_headers,
                                                   oauth_credentials)

        for key, value in headers.items():
            request.setRawHeader(key, value)

        post_buffer = QBuffer()
        post_buffer.setData(post_content)
        try:
            result = yield self._perform_request(request, method, post_buffer)
        except ProxyUnauthorizedError as e:
            app_proxy = QNetworkProxy.applicationProxy()
            proxy_host = app_proxy.hostName() if app_proxy else "proxy server"
            got_creds = yield self.request_proxy_auth_credentials(
                                            proxy_host, self.proxy_retry)
            if got_creds:
                self.proxy_retry = True
                result = yield self.request(iri, method, extra_headers,
                                            oauth_credentials, post_content)
            else:
                excp = WebClientError('Proxy creds needed.', e)
                defer.returnValue(excp)
        defer.returnValue(result)

    def _handle_authentication(self, reply, authenticator):
        """The reply needs authentication."""
        if authenticator.user() != self.username:
            authenticator.setUser(self.username)
        if authenticator.password() != self.password:
            authenticator.setPassword(self.password)

    def _handle_finished(self, reply):
        """The reply has finished processing."""
        assert reply in self.replies
        d = self.replies.pop(reply)
        error = reply.error()
        content = reply.readAll()
        if not error:
            headers = HeaderDict()
            for key, value in reply.rawHeaderPairs():
                headers[str(key)].append(str(value))
            response = Response(bytes(content), headers)
            d.callback(response)
        else:
            content = unicode(content)
            error_string = unicode(reply.errorString())
            logger.debug('_handle_finished error (%s,%s).', error,
                         error_string)
            if error == QNetworkReply.AuthenticationRequiredError:
                exception = UnauthorizedError(error_string, content)
            elif error == QNetworkReply.ProxyAuthenticationRequiredError:
                # we are going thru a proxy and we did not auth
                exception = ProxyUnauthorizedError(error_string, content)
            else:
                exception = WebClientError(error_string, content)
            d.errback(exception)

    def _get_certificate_details(self, cert):
        """Return an string with the details of the certificate."""
        detail_titles = {QSslCertificate.Organization: 'organization',
                         QSslCertificate.CommonName: 'common_name',
                         QSslCertificate.LocalityName: 'locality_name',
                         QSslCertificate.OrganizationalUnitName: 'unit',
                         QSslCertificate.CountryName: 'country_name',
                         QSslCertificate.StateOrProvinceName: 'state_name'}
        details = {}
        for info, title in detail_titles.items():
            details[title] = str(cert.issuerInfo(info))
        return self.format_ssl_details(details)

    def _get_certificate_host(self, cert):
        """Return the host of the cert."""
        return str(cert.issuerInfo(QSslCertificate.CommonName))

    def _handle_ssl_errors(self, reply, errors):
        """Handle the case in which we got an ssl error."""
        msg = StringIO()
        msg.write('SSL errors found; url: %s\n' %
                  reply.request().url().toString())
        for error in errors:
            #print error.errorString()
            msg.write('========Error=============\n%s (%s)\n' %
                     (error.errorString(), error.error()))
            msg.write('--------Cert Details------\n%s\n' %
                      self._get_certificate_details(error.certificate()))
            msg.write('==========================\n')
        logger.error(msg.getvalue())

    def force_use_proxy(self, https_settings):
        """Setup this webclient to use the given proxy settings."""
        settings = {"https": https_settings}
        proxy = build_proxy(settings)
        self._set_proxy(proxy)
        WebClient.proxy_instance = proxy

    def shutdown(self):
        """Shut down all pending requests (if possible)."""
        self.nam.deleteLater()



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Platform-specific translation functions."""

import gettext
import os
import sys

from ubuntu_kylin_sso.logger import setup_logging

logger = setup_logging('ubuntu_kylin_sso.utils.translation')


def _get_languages():
    """list of langs ordered by preference, or None for gettext defaults."""
    if sys.platform == 'darwin':
        from Cocoa import NSUserDefaults
        su = NSUserDefaults.standardUserDefaults()
        return su['AppleLanguages']
    else:
        if sys.platform == 'win32':
            return None
        return None


def _get_translations_data_path(fallback_path=None):
    """path to compiled translation files, or None for gettext defaults"""
    if getattr(sys, 'frozen', None) is not None:
        if sys.platform == 'darwin':
            main_app_dir = ''.join(__file__.partition('.app')[:-1])
            path = os.path.join(main_app_dir, 'Contents', 'Resources',
                                'translations')
            return path
        elif sys.platform == 'win32':
            return None         # TODO

    exists = (os.path.exists(fallback_path)
              if fallback_path else False)
    logger.debug("Using fallback translation path %r "
                 "which does %s exist." %
                 (fallback_path, "" if exists else "**NOT**"))

    return fallback_path


def get_gettext(translation_domain, fallback_path=None):
    """Get proper gettext translation function for platform and py version."""
    languages = _get_languages()
    translations_path = _get_translations_data_path(fallback_path)

    if languages is None or translations_path is None or languages[0] == 'en':
        logger.debug('Using default gettext translation search paths')

        translation = gettext.translation(translation_domain, fallback=True)
    else:
        translation = gettext.translation(translation_domain,
                                          translations_path,
                                          languages=languages[:1],
                                          fallback=True)
    if isinstance(translation, gettext.NullTranslations):
        logger.warn('Translations not found, using null translator.')
    if sys.version_info < (3,):
        return translation.ugettext
    else:
        return translation.gettext



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the oauth_headers helper function."""

from __future__ import unicode_literals

import logging
import sys
import os

from twisted.internet import defer
from twisted.web import resource
from ubuntuone.devtools.handlers import MementoHandler
from ubuntuone.devtools.testing.txwebserver import HTTPWebServer

from ubuntu_kylin_sso import utils
from ubuntu_kylin_sso.tests import TestCase

CONSTANTS_MODULE = 'ubuntu_kylin_sso.constants'
NOT_DEFINED = object()


class FakeWebclient(object):
    """A fake webclient."""

    def __init__(self):
        self.called = []

    def request(self, *args, **kwargs):
        """Save a webclient request."""
        self.called.append((args, kwargs))
        response = utils.webclient.common.Response(content="response")
        return response

    def shutdown(self):
        """Shutdown this fake webclient."""


class FakedConstantsModule(object):
    """Fake the 'ubuntuone.controlpanel.constants' module."""

    PROJECT_DIR = '/tmp/foo/bar'
    BIN_DIR = '/tmp/foo/bin'


class GetProjectDirTestCase(TestCase):
    """Test case for get_project_dir when constants module is not defined."""

    DIR_NAME = utils.DATA_SUFFIX
    DIR_CONSTANT = 'PROJECT_DIR'
    DIR_GETTER = 'get_project_dir'

    @defer.inlineCallbacks
    def setUp(self):
        yield super(GetProjectDirTestCase, self).setUp()
        self._constants = sys.modules.get(CONSTANTS_MODULE, NOT_DEFINED)
        sys.modules[CONSTANTS_MODULE] = None  # force ImportError

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        utils.logger.addHandler(self.memento)
        self.addCleanup(utils.logger.removeHandler, self.memento)

        self.get_dir = getattr(utils, self.DIR_GETTER)

    @defer.inlineCallbacks
    def tearDown(self):
        if self._constants is not NOT_DEFINED:
            sys.modules[CONSTANTS_MODULE] = self._constants
        else:
            sys.modules.pop(CONSTANTS_MODULE)
        yield super(GetProjectDirTestCase, self).tearDown()

    def test_get_dir_relative(self):
        """The relative path for the data directory is correctly retrieved."""
        module = utils.os.path.dirname(utils.__file__)
        rel_data = utils.os.path.join(module,
                                      utils.os.path.pardir,
                                      utils.os.path.pardir,
                                      self.DIR_NAME)
        expected_dir = utils.os.path.abspath(rel_data)

        # ensure expected_path exists at os level
        self.patch(utils.os.path, 'exists', lambda path: path == expected_dir)

        result = self.get_dir()
        self.assertEqual(expected_dir, result)

    def test_get_dir_none_exists(self):
        """No data directory exists, return None and log as error."""
        self.patch(utils.os.path, 'exists', lambda path: False)
        sys.modules[CONSTANTS_MODULE] = None

        self.assertRaises(AssertionError, self.get_dir)
        msg = 'get_dir: can not build a valid path.'
        self.assertTrue(self.memento.check_error(msg))


class GetProjectDirWithConstantsTestCase(GetProjectDirTestCase):
    """Test case for get_dir when constants module is defined."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(GetProjectDirWithConstantsTestCase, self).setUp()
        self.patch(utils.os.path, 'exists', lambda path: False)
        self._constants = sys.modules.get(CONSTANTS_MODULE, NOT_DEFINED)
        sys.modules[CONSTANTS_MODULE] = FakedConstantsModule()

    def test_get_dir(self):
        """If the constants.py module exists, use PROJECT_DIR from it."""
        result = self.get_dir()
        expected = getattr(sys.modules[CONSTANTS_MODULE], self.DIR_CONSTANT)
        self.assertEqual(expected, result)


class GetBinDirTestCase(GetProjectDirTestCase):
    """Test case for get_bin_dir when constants module is not defined."""

    DIR_NAME = utils.BIN_SUFFIX
    DIR_CONSTANT = 'BIN_DIR'
    DIR_GETTER = 'get_bin_dir'


class GetBinDirWithConstantsTestCase(GetProjectDirWithConstantsTestCase):
    """Test case for get_bin_dir when constants module is defined."""

    DIR_NAME = utils.BIN_SUFFIX
    DIR_CONSTANT = 'BIN_DIR'
    DIR_GETTER = 'get_bin_dir'


class GetDataFileTestCase(TestCase):
    """Test cases for get_data_file."""

    def test_get_data_file(self):
        """The path for a data file is correctly retrieved."""
        dummy_dir = '/yadda/yadda'
        dummy_file = 'test.png'
        self.patch(utils, 'get_project_dir', lambda: dummy_dir)
        result = utils.get_data_file(dummy_file)
        expected = utils.os.path.join(dummy_dir, dummy_file)
        self.assertEqual(expected, result)


class GetCertDirTestCase(TestCase):
    """Test determining the cert location."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(GetCertDirTestCase, self).setUp()

    def test_win(self):
        """Test geting a path when Common AppData is defined."""
        self.patch(utils, "__file__",
                   os.path.join("path", "to", "ubuntu_kylin_sso",
                                "utils", "__init__.py"))
        self.patch(sys, "platform", "win32")
        path = utils.get_cert_dir()
        self.assertEqual(path, os.path.join("path", "to", "data"))

    def test_darwin_frozen(self):
        """Test that we get a path with .app in it on frozen darwin."""
        self.patch(sys, "platform", "darwin")
        sys.frozen = "macosx-app"
        self.addCleanup(delattr, sys, "frozen")
        self.patch(utils, "__file__",
                   os.path.join("path", "to", "Main.app", "ignore"))
        path = utils.get_cert_dir()
        self.assertEqual(path, os.path.join("path", "to", "Main.app",
                                            "Contents", "Resources"))

    def test_darwin_unfrozen(self):
        """Test that we get a source-relative path on unfrozen darwin."""
        self.patch(sys, "platform", "darwin")
        self.patch(utils, "__file__",
                   os.path.join("path", "to", "ubuntuone",
                                "utils", "__init__.py"))
        path = utils.get_cert_dir()
        self.assertEqual(path, os.path.join("path", "to", "data"))

    def test_linux(self):
        """Test that linux gets the right path."""
        self.patch(sys, "platform", "linux2")
        path = utils.get_cert_dir()
        self.assertEqual(path, "/etc/ssl/certs")


class RootResource(resource.Resource):
    """A root resource that logs the number of calls."""

    isLeaf = True

    def __init__(self, *args, **kwargs):
        """Initialize this fake instance."""
        resource.Resource.__init__(self, *args, **kwargs)
        self.count = 0
        self.request_headers = []

    # pylint: disable=C0103
    def render_HEAD(self, request):
        """Increase the counter on each render."""
        self.count += 1
        self.request_headers.append(request.requestHeaders)
        return ""


class MockWebServer(HTTPWebServer):
    """A mock webserver for testing."""

    def __init__(self):
        """Start up this instance."""
        # pylint: disable=E1101
        root = RootResource()
        super(MockWebServer, self).__init__(root)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT AN WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Integration tests for the proxy-enabled webclient."""

import logging
import os
import shutil
import sys

try:
    # pylint: disable=E0611,F0401
    from urllib.parse import (urlencode, unquote, urlparse,
                              urljoin, parse_qsl)
    # pylint: enable=E0611,F0401
except ImportError:
    from urllib import urlencode, unquote
    from urlparse import urlparse, urljoin, parse_qsl

from OpenSSL import crypto
from socket import gethostname
from twisted.cred import checkers, portal
from twisted.internet import defer
from twisted.web import guard, http, resource
from ubuntuone.devtools.handlers import MementoHandler
from ubuntuone.devtools.testcases import TestCase, skipIfOS
from ubuntuone.devtools.testcases.squid import SquidTestCase
from ubuntuone.devtools.testing.txwebserver import (
    HTTPWebServer,
    HTTPSWebServer,
)


from ubuntu_kylin_sso import (
    keyring,
    EXCEPTION_RAISED,
    USER_SUCCESS,
    USER_CANCELLATION,
)
from ubuntu_kylin_sso.utils import webclient
from ubuntu_kylin_sso.utils.ui import SSL_DETAILS_TEMPLATE
from ubuntu_kylin_sso.utils.webclient import gsettings, txweb
from ubuntu_kylin_sso.utils.webclient.common import (
    BaseWebClient,
    HeaderDict,
    UnauthorizedError,
    WebClientError,
)

ANY_VALUE = object()
SAMPLE_KEY = "result"
SAMPLE_VALUE = "sample result"
SAMPLE_RESOURCE = '{"%s": "%s"}' % (SAMPLE_KEY, SAMPLE_VALUE)
SAMPLE_USERNAME = "peddro"
SAMPLE_PASSWORD = "cantropus"
SAMPLE_CREDENTIALS = dict(
    consumer_key="consumer key",
    consumer_secret="consumer secret",
    token="the real token",
    token_secret="the token secret",
)
SAMPLE_HEADERS = {SAMPLE_KEY: SAMPLE_VALUE}
SAMPLE_POST_PARAMS = {"param1": "value1", "param2": "value2"}
SAMPLE_JPEG_HEADER = '\xff\xd8\xff\xe0\x00\x10JFIF'

SIMPLERESOURCE = "simpleresource"
BYTEZERORESOURCE = "bytezeroresource"
POSTABLERESOURCE = "postableresource"
THROWERROR = "throwerror"
UNAUTHORIZED = "unauthorized"
HEADONLY = "headonly"
VERIFYHEADERS = "verifyheaders"
VERIFYPOSTPARAMS = "verifypostparams"
GUARDED = "guarded"
OAUTHRESOURCE = "oauthresource"

WEBCLIENT_MODULE_NAME = webclient.webclient_module().__name__


def sample_get_credentials():
    """Will return the sample credentials right now."""
    return defer.succeed(SAMPLE_CREDENTIALS)


# pylint: disable=C0103
# t.w.resource methods have freeform cased names

class SimpleResource(resource.Resource):
    """A simple web resource."""

    def render_GET(self, request):
        """Make a bit of html out of these resource's content."""
        return SAMPLE_RESOURCE


class ByteZeroResource(resource.Resource):
    """A resource that has a nul byte in the middle of it."""

    def render_GET(self, request):
        """Return the content of this resource."""
        return SAMPLE_JPEG_HEADER


class PostableResource(resource.Resource):
    """A resource that only answers to POST requests."""

    def render_POST(self, request):
        """Make a bit of html out of these resource's content."""
        return SAMPLE_RESOURCE


class HeadOnlyResource(resource.Resource):
    """A resource that fails if called with a method other than HEAD."""

    def render_HEAD(self, request):
        """Return a bit of html."""
        return "OK"


class VerifyHeadersResource(resource.Resource):
    """A resource that verifies the headers received."""

    def render_GET(self, request):
        """Make a bit of html out of these resource's content."""
        headers = request.requestHeaders.getRawHeaders(SAMPLE_KEY)
        if headers != [SAMPLE_VALUE]:
            request.setResponseCode(http.BAD_REQUEST)
            return "ERROR: Expected header not present."
        request.setHeader(SAMPLE_KEY, SAMPLE_VALUE)
        return SAMPLE_RESOURCE


class VerifyPostParameters(resource.Resource):
    """A resource that answers to POST requests with some parameters."""

    def fetch_post_args_only(self, request):
        """Fetch only the POST arguments, not the args in the url."""
        request.process = lambda: None
        request.requestReceived(request.method, request.path,
                                request.clientproto)
        return request.args

    def render_POST(self, request):
        """Verify the parameters that we've been called with."""
        post_params = self.fetch_post_args_only(request)
        expected = dict((key, [val]) for key, val
                                      in SAMPLE_POST_PARAMS.items())
        if post_params != expected:
            request.setResponseCode(http.BAD_REQUEST)
            return "ERROR: Expected arguments not present, %r != %r" % (
                    post_params, expected)
        return SAMPLE_RESOURCE


class SimpleRealm(object):
    """The same simple resource for all users."""

    def requestAvatar(self, avatarId, mind, *interfaces):
        """The avatar for this user."""
        if resource.IResource in interfaces:
            return (resource.IResource, SimpleResource(), lambda: None)
        raise NotImplementedError()


class OAuthCheckerResource(resource.Resource):
    """A resource that verifies the request was oauth signed."""

    def render_GET(self, request):
        """Make a bit of html out of these resource's content."""
        header = request.requestHeaders.getRawHeaders("Authorization")[0]
        if header.startswith("OAuth "):
            return SAMPLE_RESOURCE
        request.setResponseCode(http.BAD_REQUEST)
        return "ERROR: Expected OAuth header not present."


def get_root_resource():
    """Get the root resource with all the children."""
    root = resource.Resource()
    root.putChild(SIMPLERESOURCE, SimpleResource())
    root.putChild(BYTEZERORESOURCE, ByteZeroResource())
    root.putChild(POSTABLERESOURCE, PostableResource())

    root.putChild(THROWERROR, resource.NoResource())

    unauthorized_resource = resource.ErrorPage(http.UNAUTHORIZED,
                                               "Unauthorized", "Unauthorized")
    root.putChild(UNAUTHORIZED, unauthorized_resource)
    root.putChild(HEADONLY, HeadOnlyResource())
    root.putChild(VERIFYHEADERS, VerifyHeadersResource())
    root.putChild(VERIFYPOSTPARAMS, VerifyPostParameters())
    root.putChild(OAUTHRESOURCE, OAuthCheckerResource())

    db = checkers.InMemoryUsernamePasswordDatabaseDontUse()
    db.addUser(SAMPLE_USERNAME, SAMPLE_PASSWORD)
    test_portal = portal.Portal(SimpleRealm(), [db])
    cred_factory = guard.BasicCredentialFactory("example.org")
    guarded_resource = guard.HTTPAuthSessionWrapper(test_portal,
                                                        [cred_factory])
    root.putChild(GUARDED, guarded_resource)
    return root


class HTTPMockWebServer(HTTPWebServer):
    """A mock webserver for the webclient tests."""

    def __init__(self):
        """Create a new instance."""
        root = get_root_resource()
        super(HTTPMockWebServer, self).__init__(root)


class HTTPSMockWebServer(HTTPSWebServer):
    """A mock webserver for the webclient tests."""

    def __init__(self, ssl_settings):
        """Create a new instance."""
        root = get_root_resource()
        super(HTTPSMockWebServer, self).__init__(root, ssl_settings)


class ModuleSelectionTestCase(TestCase):
    """Test the functions to choose the qtnet or libsoup backend."""

    def test_is_qt4reactor_installed_not_installed(self):
        """When the qt4reactor is not installed, it returns false."""
        self.patch(sys, "modules", {})
        self.assertFalse(webclient.is_qt4reactor_installed())

    def test_is_qt4reactor_installed_installed_core(self):
        """When the qt4reactor is installed, it returns true."""
        self.assertTrue(webclient.is_qt4reactor_installed())

    def assert_module_name(self, module, expected_name):
        """Check the name of a given module."""
        module_filename = os.path.basename(module.__file__)
        module_name = os.path.splitext(module_filename)[0]
        self.assertEqual(module_name, expected_name)

    def test_webclient_module_qtnetwork(self):
        """Test the module name for the qtnetwork case."""
        self.patch(webclient, "is_qt4reactor_installed", lambda: True)
        module = webclient.webclient_module()
        self.assert_module_name(module, "qtnetwork")

    def test_webclient_module_libsoup(self):
        """Test the module name for the libsoup case."""
        self.patch(webclient, "is_qt4reactor_installed", lambda: False)
        module = webclient.webclient_module()
        self.assert_module_name(module, "libsoup")


class WebClientTestCase(TestCase):
    """Test for the webclient."""

    timeout = 1
    webclient_factory = webclient.webclient_factory

    @defer.inlineCallbacks
    def setUp(self):
        yield super(WebClientTestCase, self).setUp()
        self.wc = self.webclient_factory()
        self.addCleanup(self.wc.shutdown)
        self.ws = HTTPMockWebServer()
        self.ws.start()
        self.addCleanup(self.ws.stop)
        self.base_iri = self.ws.get_iri()

    @defer.inlineCallbacks
    def test_request_takes_an_iri(self):
        """Passing a non-unicode iri fails."""
        d = self.wc.request(bytes(self.base_iri + SIMPLERESOURCE))
        yield self.assertFailure(d, TypeError)

    @defer.inlineCallbacks
    def test_get_iri(self):
        """Passing in a unicode iri works fine."""
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assertEqual(SAMPLE_RESOURCE, result.content)

    @defer.inlineCallbacks
    def test_get_iri_error(self):
        """The errback is called when there's some error."""
        yield self.assertFailure(self.wc.request(self.base_iri + THROWERROR),
                                 WebClientError)

    @defer.inlineCallbacks
    def test_zero_byte_in_content(self):
        """Test a reply with a nul byte in the middle of it."""
        result = yield self.wc.request(self.base_iri + BYTEZERORESOURCE)
        self.assertEqual(SAMPLE_JPEG_HEADER, result.content)

    @defer.inlineCallbacks
    def test_post(self):
        """Test a post request."""
        result = yield self.wc.request(self.base_iri + POSTABLERESOURCE,
                                       method="POST")
        self.assertEqual(SAMPLE_RESOURCE, result.content)

    @defer.inlineCallbacks
    def test_post_with_args(self):
        """Test a post request with arguments."""
        args = urlencode(SAMPLE_POST_PARAMS)
        iri = self.base_iri + VERIFYPOSTPARAMS + "?" + args
        headers = {
            "content-type": "application/x-www-form-urlencoded",
        }
        result = yield self.wc.request(iri, method="POST",
                                      extra_headers=headers, post_content=args)
        self.assertEqual(SAMPLE_RESOURCE, result.content)

    @defer.inlineCallbacks
    def test_unauthorized(self):
        """Detect when a request failed with the UNAUTHORIZED http code."""
        yield self.assertFailure(self.wc.request(self.base_iri + UNAUTHORIZED),
                                 UnauthorizedError)

    @defer.inlineCallbacks
    def test_method_head(self):
        """The HTTP method is used."""
        result = yield self.wc.request(self.base_iri + HEADONLY, method="HEAD")
        self.assertEqual("", result.content)

    @defer.inlineCallbacks
    def test_send_extra_headers(self):
        """The extra_headers are sent to the server."""
        result = yield self.wc.request(self.base_iri + VERIFYHEADERS,
                                       extra_headers=SAMPLE_HEADERS)
        self.assertIn(SAMPLE_KEY, result.headers)
        self.assertEqual(result.headers[SAMPLE_KEY], [SAMPLE_VALUE])

    @defer.inlineCallbacks
    def test_send_basic_auth(self):
        """The basic authentication headers are sent."""
        other_wc = self.webclient_factory(username=SAMPLE_USERNAME,
                                          password=SAMPLE_PASSWORD)
        self.addCleanup(other_wc.shutdown)
        result = yield other_wc.request(self.base_iri + GUARDED)
        self.assertEqual(SAMPLE_RESOURCE, result.content)

    @defer.inlineCallbacks
    def test_send_basic_auth_wrong_credentials(self):
        """Wrong credentials returns a webclient error."""
        other_wc = self.webclient_factory(username=SAMPLE_USERNAME,
                                          password="wrong password!")
        self.addCleanup(other_wc.shutdown)
        yield self.assertFailure(other_wc.request(self.base_iri + GUARDED),
                                 UnauthorizedError)

    @defer.inlineCallbacks
    def test_request_is_oauth_signed(self):
        """The request is oauth signed."""
        tsc = self.wc.get_timestamp_checker()
        self.patch(tsc, "get_faithful_time", lambda: defer.succeed('1'))
        result = yield self.wc.request(self.base_iri + OAUTHRESOURCE,
                                       oauth_credentials=SAMPLE_CREDENTIALS)
        self.assertEqual(SAMPLE_RESOURCE, result.content)

    @defer.inlineCallbacks
    def test_oauth_signing_uses_timestamp(self):
        """OAuth signing uses the timestamp."""
        called = []

        def fake_get_faithful_time():
            """A fake get_timestamp"""
            called.append(True)
            return defer.succeed('1')

        tsc = self.wc.get_timestamp_checker()
        self.patch(tsc, "get_faithful_time", fake_get_faithful_time)
        yield self.wc.request(self.base_iri + OAUTHRESOURCE,
                              oauth_credentials=SAMPLE_CREDENTIALS)
        self.assertTrue(called, "The timestamp must be retrieved.")

    @defer.inlineCallbacks
    def test_returned_content_are_bytes(self):
        """The returned content are bytes."""
        tsc = self.wc.get_timestamp_checker()
        self.patch(tsc, "get_faithful_time", lambda: defer.succeed('1'))
        result = yield self.wc.request(self.base_iri + OAUTHRESOURCE,
                                       oauth_credentials=SAMPLE_CREDENTIALS)
        self.assertTrue(isinstance(result.content, bytes),
                        "The type of %r must be bytes" % result.content)

    @defer.inlineCallbacks
    def test_webclienterror_not_string(self):
        """The returned exception contains unicode data."""
        deferred = self.wc.request(self.base_iri + THROWERROR)
        failure = yield self.assertFailure(deferred, WebClientError)
        for error in failure.args:
            self.assertTrue(isinstance(error, basestring))


class FakeSavingReactor(object):
    """A fake reactor that saves connection attempts."""

    def __init__(self):
        """Initialize this fake instance."""
        self.connections = []

    def connectTCP(self, host, port, factory, *args):
        """Fake the connection."""
        self.connections.append((host, port, args))
        factory.response_headers = {}
        factory.deferred = defer.succeed("response content")

    def connectSSL(self, host, port, factory, *args):
        """Fake the connection."""
        self.connections.append((host, port, args))
        factory.response_headers = {}
        factory.deferred = defer.succeed("response content")


class TxWebClientTestCase(WebClientTestCase):
    """Test case for txweb."""

    webclient_factory = txweb.WebClient

    @defer.inlineCallbacks
    def setUp(self):
        """Set the diff tests."""
        # delay import, otherwise a default reactor gets installed
        from twisted.web import client
        self.factory = client.HTTPClientFactory
        # set the factory to be used
        # Hook the server's buildProtocol to make the protocol instance
        # accessible to tests and ensure that the reactor is clean!
        build_protocol = self.factory.buildProtocol
        self.serverProtocol = None

        def remember_protocol_instance(my_self, addr):
            """Remember the protocol used in the test."""
            protocol = build_protocol(my_self, addr)
            self.serverProtocol = protocol
            on_connection_lost = defer.Deferred()
            connection_lost = protocol.connectionLost

            def defer_connection_lost(protocol, *a):
                """Lost connection."""
                if not on_connection_lost.called:
                    on_connection_lost.callback(None)
                connection_lost(protocol, *a)

            self.patch(protocol, 'connectionLost', defer_connection_lost)

            def cleanup():
                """Clean the connection."""
                if self.serverProtocol.transport is not None:
                    self.serverProtocol.transport.loseConnection()
                return on_connection_lost

            self.addCleanup(cleanup)
            return protocol

        self.factory.buildProtocol = remember_protocol_instance
        self.addCleanup(self.set_build_protocol, build_protocol)
        txweb.WebClient.client_factory = self.factory

        yield super(TxWebClientTestCase, self).setUp()

    def set_build_protocol(self, method):
        """Set the method back."""
        self.factory.buildProtocol = method


class TxWebClientReactorReplaceableTestCase(TestCase):
    """In the txweb client the reactor is replaceable."""

    timeout = 3
    FAKE_HOST = u"fake"
    FAKE_IRI_TEMPLATE = u"%%s://%s/fake_page" % FAKE_HOST

    @defer.inlineCallbacks
    def _test_replaceable_reactor(self, iri):
        """The reactor can be replaced with the tunnel client."""
        fake_reactor = FakeSavingReactor()
        wc = txweb.WebClient(fake_reactor)
        _response = yield wc.request(iri)
        assert(_response)
        host, _port, _args = fake_reactor.connections[0]
        self.assertEqual(host, self.FAKE_HOST)

    def test_replaceable_reactor_http(self):
        """Test the replaceable reactor with an http iri."""
        return self._test_replaceable_reactor(self.FAKE_IRI_TEMPLATE % "http")

    def test_replaceable_reactor_https(self):
        """Test the replaceable reactor with an https iri."""
        return self._test_replaceable_reactor(self.FAKE_IRI_TEMPLATE % "https")


class TimestampCheckerTestCase(TestCase):
    """Tests for the timestampchecker classmethod."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this testcase."""
        yield super(TimestampCheckerTestCase, self).setUp()
        self.wc = webclient.webclient_factory()
        self.patch(self.wc.__class__, "timestamp_checker", None)

    def test_timestamp_checker_has_the_same_class_as_the_creator(self):
        """The TimestampChecker has the same class."""
        tsc = self.wc.get_timestamp_checker()
        self.assertEqual(tsc.webclient_class, self.wc.__class__)

    def test_timestamp_checker_is_the_same_for_all_webclients(self):
        """The TimestampChecker is the same for all webclients."""
        tsc1 = self.wc.get_timestamp_checker()
        wc2 = webclient.webclient_factory()
        tsc2 = wc2.get_timestamp_checker()
        # pylint: disable=E1101
        self.assertIs(tsc1, tsc2)


class BasicProxyTestCase(SquidTestCase):
    """Test that the proxy works at all."""

    timeout = 3

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BasicProxyTestCase, self).setUp()
        self.ws = HTTPMockWebServer()
        self.ws.start()
        self.addCleanup(self.ws.stop)
        self.base_iri = self.ws.get_iri()
        self.wc = webclient.webclient_factory()
        self.addCleanup(self.wc.shutdown)

    def assert_header_contains(self, headers, expected):
        """One of the headers matching key must contain a given value."""
        self.assertTrue(any(expected in value for value in headers))

    @defer.inlineCallbacks
    def test_anonymous_proxy_is_used(self):
        """The anonymous proxy is used by the webclient."""
        settings = self.get_nonauth_proxy_settings()
        self.wc.force_use_proxy(settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    @defer.inlineCallbacks
    def test_authenticated_proxy_is_used(self):
        """The authenticated proxy is used by the webclient."""
        settings = self.get_auth_proxy_settings()
        self.wc.force_use_proxy(settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    @defer.inlineCallbacks
    def test_auth_proxy_is_used_creds_requested(self):
        """The authenticated proxy is used by the webclient."""
        settings = self.get_auth_proxy_settings()
        partial_settings = dict(host=settings['host'], port=settings['port'])

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            self.wc.proxy_username = settings['username']
            self.wc.proxy_password = settings['password']
            return defer.succeed(True)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

        self.wc.force_use_proxy(partial_settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    @defer.inlineCallbacks
    def test_auth_proxy_is_requested_creds_bad_details(self):
        """Test using wrong credentials with the proxy."""
        settings = self.get_auth_proxy_settings()
        wrong_settings = dict(host=settings['host'], port=settings['port'],
                              username=settings['password'],
                              password=settings['username'])

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            self.wc.proxy_username = settings['username']
            self.wc.proxy_password = settings['password']
            return defer.succeed(True)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

        self.wc.force_use_proxy(wrong_settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    @defer.inlineCallbacks
    def test_auth_proxy_is_requested_creds_bad_details_user(self):
        """Test using no creds and user providing the wrong ones."""
        settings = self.get_auth_proxy_settings()
        partial_settings = dict(host=settings['host'], port=settings['port'])

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            if retry:
                self.wc.proxy_username = settings['username']
                self.wc.proxy_password = settings['password']
            else:
                self.wc.proxy_username = settings['password']
                self.wc.proxy_password = settings['username']
            return defer.succeed(True)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

        self.wc.force_use_proxy(partial_settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    @defer.inlineCallbacks
    def test_auth_proxy_is_requested_creds_bad_details_everywhere(self):
        """Test when we pass the wrong settings and get the wrong settings."""
        settings = self.get_auth_proxy_settings()
        wrong_settings = dict(host=settings['host'], port=settings['port'],
                              username=settings['password'],
                              password=settings['username'])

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            if retry:
                self.wc.proxy_username = settings['username']
                self.wc.proxy_password = settings['password']
            else:
                self.wc.proxy_username = settings['password']
                self.wc.proxy_password = settings['username']
            return defer.succeed(True)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

        self.wc.force_use_proxy(wrong_settings)
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")

    @skipIfOS('linux2',
              'LP: #1111880 - ncsa_auth crashing for auth proxy tests.')
    def test_auth_proxy_is_requested_user_cancels(self):
        """Test when the user cancels the creds dialog."""
        settings = self.get_auth_proxy_settings()
        partial_settings = dict(host=settings['host'], port=settings['port'])

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            return defer.succeed(False)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

        self.wc.force_use_proxy(partial_settings)
        self.failUnlessFailure(self.wc.request(self.base_iri + SIMPLERESOURCE),
                               WebClientError)

    if WEBCLIENT_MODULE_NAME.endswith(".txweb"):
        reason = "txweb does not support proxies."
        test_anonymous_proxy_is_used.skip = reason
        test_authenticated_proxy_is_used.kip = reason
        test_auth_proxy_is_used_creds_requested.skip = reason
        test_auth_proxy_is_requested_creds_bad_details.skip = reason
        test_auth_proxy_is_requested_creds_bad_details_user.skip = reason
        test_auth_proxy_is_requested_creds_bad_details_everywhere.skip = reason
        test_auth_proxy_is_requested_user_cancels.skip = reason


class HeaderDictTestCase(TestCase):
    """Tests for the case insensitive header dictionary."""

    def test_constructor_handles_keys(self):
        """The constructor handles case-insensitive keys."""
        hd = HeaderDict({"ClAvE": "value"})
        self.assertIn("clave", hd)

    def test_can_set_get_items(self):
        """The item is set/getted."""
        hd = HeaderDict()
        hd["key"] = "value"
        hd["KEY"] = "value2"
        self.assertEqual(hd["key"], "value2")

    def test_can_test_presence(self):
        """The presence of an item is found."""
        hd = HeaderDict()
        self.assertNotIn("cLaVe", hd)
        hd["CLAVE"] = "value1"
        self.assertIn("cLaVe", hd)
        del(hd["cLAVe"])
        self.assertNotIn("cLaVe", hd)


class OAuthPlainTextTestCase(TestCase):
    """Test for the oauth signing code using PLAINTEXT."""

    oauth_sign = "PLAINTEXT"
    sample_method = "GET"
    sample_url = "http://one.ubuntu.com/"
    sample_params = {'next': 'yadda-yadda-yo', 'foo': 'naranja fanta'}
    timestamp = '1'
    expected_params = {
        "oauth_timestamp": str(timestamp),
        "oauth_consumer_key": SAMPLE_CREDENTIALS["consumer_key"],
        "oauth_token": SAMPLE_CREDENTIALS["token"],
        "oauth_nonce": ANY_VALUE,
        "oauth_signature": ANY_VALUE,
    }

    @defer.inlineCallbacks
    def setUp(self):
        yield super(OAuthPlainTextTestCase, self).setUp()
        self.expected_params["oauth_signature_method"] = self.oauth_sign
        oauth_sign_plain = self.oauth_sign == "PLAINTEXT"
        self.wc = BaseWebClient(oauth_sign_plain=oauth_sign_plain)

    def parse_oauth_header(self, header):
        """Parse an oauth header into a tuple of (method, params_dict)."""
        params = {}

        method, params_string = header.split(" ", 1)
        for p in params_string.split(","):
            k, v = p.strip().split("=")
            params[k] = unquote(v[1:-1])

        return method, params

    def assert_headers_correct(self, headers):
        """Check that 'headers' match expected_params."""
        method, params = self.parse_oauth_header(headers["Authorization"])

        self.assertEqual(method, "OAuth")

        for k, expected_value in self.expected_params.items():
            self.assertIn(k, params)
            if expected_value is not ANY_VALUE:
                self.assertEqual(params[k], expected_value)

    @defer.inlineCallbacks
    def test_build_oauth_headers(self):
        """Check that the oauth headers are properly built."""

        def get_timestamp():
            """Return the timestamp."""
            return defer.succeed(self.timestamp)

        self.patch(self.wc, 'get_timestamp', get_timestamp)

        headers = yield self.wc.build_request_headers(self.sample_url,
                                         method=self.sample_method,
                                         oauth_credentials=SAMPLE_CREDENTIALS)
        self.assert_headers_correct(headers)

    def test_build_oauth_request(self, params=None):
        """Check that the oauth request are properly built."""
        url, signed_headers, body = self.wc.build_oauth_request(
            self.sample_method,
            self.sample_url, SAMPLE_CREDENTIALS, self.timestamp,
            parameters=params, as_query=False)

        self.assert_headers_correct(signed_headers)
        self.assertEqual(url, self.sample_url)

    @defer.inlineCallbacks
    def test_build_signed_iri(self, params=None):
        """Check that the oauth signed iri is properly built."""
        self.patch(self.wc, 'get_timestamp', lambda: self.timestamp)
        iri = u'http://foo.bar.baz/ñandú'
        uri = self.wc.iri_to_uri(iri)
        # request will be a 3-tuple of the url, signed headers, and body.
        url, signed_headers, body = self.wc.build_oauth_request(
            self.sample_method,
            uri, SAMPLE_CREDENTIALS,
            self.timestamp, parameters=params)
        scheme, netloc, path, params, query, _ = urlparse(url)
        expected = dict(parse_qsl(query))

        result = yield self.wc.build_signed_iri(
            iri, SAMPLE_CREDENTIALS, parameters=params)
        scheme, netloc, path, params, query, _ = urlparse(result)
        self.assertEqual(urljoin(scheme + '://' + netloc, path), uri)
        actual = dict(parse_qsl(query))

        for i in ('oauth_nonce', 'oauth_signature'):
            expected.pop(i)
            actual.pop(i)
        self.assertEqual(expected, actual)

    def test_build_signed_iri_with_params(self, params=None):
        """Check that the oauth signed iri is properly built with params."""
        return self.test_build_signed_iri(params=self.sample_params)

    def patch_staticmethod(self, src, attr_name, replacement):
        """Alternate patch because self.patch fails restoring staticmethods."""
        original = getattr(src, attr_name)
        setattr(src, attr_name, staticmethod(replacement))
        self.addCleanup(setattr, src, attr_name, staticmethod(original))


class OAuthHmacSha1TestCase(OAuthPlainTextTestCase):
    """Test for the oauth signing code using HMAC-SHA1."""

    oauth_sign = "HMAC-SHA1"


class FakeKeyring(object):
    """A fake keyring."""

    def __init__(self, creds):
        """A fake keyring."""
        self.creds = creds

    def __call__(self):
        """Fake instance callable."""
        return self

    def get_credentials(self, domain):
        """A fake get_credentials."""
        if isinstance(self.creds, Exception):
            return defer.fail(self.creds)
        return defer.succeed(self.creds)


class RequestProxyAuthTestCase(TestCase):
    """Test the spawn of the creds dialog."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(RequestProxyAuthTestCase, self).setUp()
        self.wc = webclient.webclient_factory()
        self.addCleanup(self.wc.shutdown)
        self.domain = 'domain'
        self.retry = False
        self.creds = dict(username='username', password='password')

        self.keyring = FakeKeyring(self.creds)
        self.patch(keyring, 'Keyring', self.keyring)

        self.spawn_return_code = USER_SUCCESS

        def fake_spawn_process(args):
            """Fake spawning a process."""
            if isinstance(self.spawn_return_code, Exception):
                return defer.fail(self.spawn_return_code)
            return defer.succeed(self.spawn_return_code)

        self.patch(webclient.common, 'spawn_program', fake_spawn_process)

    def test_spawn_error(self):
        """Test the case when we cannot spawn the process."""
        self.spawn_return_code = Exception()
        self.failUnlessFailure(self.wc.request_proxy_auth_credentials(
                self.domain, True), WebClientError)

    @defer.inlineCallbacks
    def test_creds_acquired(self):
        """Test the case in which we do get the creds."""
        got_creds = yield self.wc.request_proxy_auth_credentials(
            self.domain, self.retry)
        self.assertTrue(got_creds, 'Return true when creds are present.')
        self.assertEqual(self.wc.proxy_username, self.creds['username'])
        self.assertEqual(self.wc.proxy_password, self.creds['password'])

    def test_creds_acquired_keyring_error(self):
        """Test the case in which we cannot access the keyring."""
        self.keyring.creds = Exception()
        self.failUnlessFailure(self.wc.request_proxy_auth_credentials(
                self.domain, self.retry), WebClientError)

    @defer.inlineCallbacks
    def test_creds_none(self):
        """Test the case in which we got None from the keyring."""
        self.keyring.creds = None
        got_creds = yield self.wc.request_proxy_auth_credentials(self.domain,
                                                           self.retry)
        self.assertFalse(got_creds, 'Return false when creds are not present.')

    @defer.inlineCallbacks
    def test_user_cancelation(self):
        """Test the case in which the user cancels."""
        self.spawn_return_code = USER_CANCELLATION
        got_creds = yield self.wc.request_proxy_auth_credentials(self.domain,
                                                           self.retry)
        self.assertTrue(got_creds, 'Return true when user cancels.')

    @defer.inlineCallbacks
    def test_exception_error(self):
        """Test the case in which something bad happened."""
        self.spawn_return_code = EXCEPTION_RAISED
        got_creds = yield self.wc.request_proxy_auth_credentials(self.domain,
                                                           self.retry)
        self.assertTrue(got_creds, 'Return true when user cancels.')


class BaseSSLTestCase(SquidTestCase):
    """Base test that allows to use ssl connections."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the diff tests."""
        yield super(BaseSSLTestCase, self).setUp()
        self.cert_dir = os.path.join(self.tmpdir, 'cert')
        self.cert_details = dict(organization='Canonical',
                                 common_name=gethostname(),
                                 locality_name='London',
                                 unit='Ubuntu One',
                                 country_name='UK',
                                 state_name='London',)
        self.ssl_settings = self._generate_self_signed_certificate(
            self.cert_dir,
            self.cert_details)
        self.addCleanup(self._clean_ssl_certificate_files)

        self.ws = HTTPSMockWebServer(self.ssl_settings)
        self.ws.start()
        self.addCleanup(self.ws.stop)
        self.base_iri = self.ws.get_iri()

    def _clean_ssl_certificate_files(self):
        """Remove the certificate files."""
        if os.path.exists(self.cert_dir):
            shutil.rmtree(self.cert_dir)

    def _generate_self_signed_certificate(self, cert_dir, cert_details):
        """Generate the required SSL certificates."""
        if not os.path.exists(cert_dir):
            os.makedirs(cert_dir)
        cert_path = os.path.join(cert_dir, 'cert.crt')
        key_path = os.path.join(cert_dir, 'cert.key')

        if os.path.exists(cert_path):
            os.unlink(cert_path)
        if os.path.exists(key_path):
            os.unlink(key_path)

        # create a key pair
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.get_subject().C = cert_details['country_name']
        cert.get_subject().ST = cert_details['state_name']
        cert.get_subject().L = cert_details['locality_name']
        cert.get_subject().O = cert_details['organization']
        cert.get_subject().OU = cert_details['unit']
        cert.get_subject().CN = cert_details['common_name']
        cert.set_serial_number(1000)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(key)
        cert.sign(key, 'sha1')

        with open(cert_path, 'wt') as fd:
            fd.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

        with open(key_path, 'wt') as fd:
            fd.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))

        return dict(key=key_path, cert=cert_path)


class CorrectProxyTestCase(BaseSSLTestCase):
    """Test the interaction with a SSL enabled proxy."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the tests."""
        yield super(CorrectProxyTestCase, self).setUp()

        # fake the gsettings to have diff settings for https and http
        http_settings = self.get_auth_proxy_settings()

        # remember so that we can use them in the creds request
        proxy_username = http_settings['username']
        proxy_password = http_settings['password']

        # delete the username and password so that we get a 407 for testing
        del http_settings['username']
        del http_settings['password']

        https_settings = self.get_nonauth_proxy_settings()

        proxy_settings = dict(http=http_settings, https=https_settings)
        self.patch(gsettings, "get_proxy_settings", lambda: proxy_settings)

        self.wc = webclient.webclient_factory()
        self.addCleanup(self.wc.shutdown)

        self.called = []

        def fake_creds_request(domain, retry):
            """Fake user interaction."""
            self.called.append('request_proxy_auth_credentials')
            self.wc.proxy_username = proxy_username
            self.wc.proxy_password = proxy_password
            return defer.succeed(True)

        self.patch(self.wc, 'request_proxy_auth_credentials',
                   fake_creds_request)

    def assert_header_contains(self, headers, expected):
        """One of the headers matching key must contain a given value."""
        self.assertTrue(any(expected in value for value in headers))

    @defer.inlineCallbacks
    def test_https_request(self):
        """Test using the correct proxy for the ssl request.

        In order to assert that the correct proxy is used we expect not to call
        the auth dialog since we set the https proxy not to use the auth proxy
        and to fail because we are reaching a https page with bad self-signed
        certs.
        """
        # we fail due to the fake ssl cert
        yield self.failUnlessFailure(self.wc.request(
                                     self.base_iri + SIMPLERESOURCE),
                                     WebClientError)
        # https requests do not use the auth proxy therefore called should be
        # empty. This asserts that we are using the correct settings for the
        # request.
        self.assertEqual([], self.called)

    @defer.inlineCallbacks
    def test_http_request(self):
        """Test using the correct proxy for the plain request.

        This tests does the opposite to the https tests. We did set the auth
        proxy for the http request therefore we expect the proxy dialog to be
        used and not to get an error since we are not visiting a https with bad
        self-signed certs.
        """
        # we do not fail since we are not going to the https page
        result = yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        self.assert_header_contains(result.headers["Via"], "squid")
        # assert that we did go through the auth proxy
        self.assertIn('request_proxy_auth_credentials', self.called)

    if WEBCLIENT_MODULE_NAME.endswith(".txweb"):
        reason = 'Multiple proxy settings is not supported.'
        test_https_request.skip = reason
        test_http_request.skip = reason

    if WEBCLIENT_MODULE_NAME.endswith(".libsoup"):
        reason = 'Hard to test since we need to fully mock gsettings.'
        test_https_request.skip = reason
        test_http_request.skip = reason

    if WEBCLIENT_MODULE_NAME.endswith(".qtnetwork"):
        reason = ('Updating proxy settings is not well support due to bug'
                  ' QTBUG-14850: https://bugreports.qt-project.org/'
                  'browse/QTBUG-14850')
        test_https_request.skip = reason
        test_http_request.skip = reason


class SSLTestCase(BaseSSLTestCase):
    """Test error handling when dealing with ssl."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the diff tests."""
        yield super(SSLTestCase, self).setUp()

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        logger = webclient.webclient_module().logger
        logger.addHandler(self.memento)
        self.addCleanup(logger.removeHandler, self.memento)

        self.wc = webclient.webclient_factory()
        self.addCleanup(self.wc.shutdown)

        self.return_code = USER_CANCELLATION
        self.called = []

        def fake_launch_ssl_dialog(client, domain, details):
            """Fake the ssl dialog."""
            self.called.append(('_launch_ssl_dialog', domain, details))
            return defer.succeed(self.return_code)

        self.patch(BaseWebClient, '_launch_ssl_dialog', fake_launch_ssl_dialog)

    @defer.inlineCallbacks
    def _assert_ssl_fail_user_accepts(self, proxy_settings=None):
        """Assert the dialog is shown in an ssl fail."""
        self.return_code = USER_SUCCESS
        if proxy_settings:
            self.wc.force_use_proxy(proxy_settings)
        yield self.wc.request(self.base_iri + SIMPLERESOURCE)
        details = SSL_DETAILS_TEMPLATE % self.cert_details
        self.assertIn(('_launch_ssl_dialog', gethostname(), details),
                      self.called)

    def test_ssl_fail_dialog_user_accepts(self):
        """Test showing the dialog and accepting."""
        self._assert_ssl_fail_user_accepts()

    def test_ssl_fail_dialog_user_accepts_via_proxy(self):
        """Test showing the dialog and accepting when using a proxy."""
        self._assert_ssl_fail_user_accepts(self.get_nonauth_proxy_settings())

    def test_ssl_fail(self):
        """Test showing the dialog and rejecting."""
        self.failUnlessFailure(self.wc.request(
                self.base_iri + SIMPLERESOURCE), WebClientError)
        self.assertNotEqual(None, self.memento.check_error('SSL errors'))

    def test_format_ssl_details(self):
        """Assert that details are correctly formatted"""
        details = SSL_DETAILS_TEMPLATE % self.cert_details
        self.assertEqual(details,
                self.wc.format_ssl_details(self.cert_details))

    if (WEBCLIENT_MODULE_NAME.endswith(".txweb") or
            WEBCLIENT_MODULE_NAME.endswith(".libsoup")):
        reason = 'SSL support has not yet been implemented.'
        test_ssl_fail_dialog_user_accepts.skip = reason
        test_ssl_fail_dialog_user_accepts_via_proxy.skip = reason
        test_ssl_fail.skip = reason



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT AN WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the proxy-enabled restful client."""

import logging

try:
    # pylint: disable=E0611,F0401
    from urllib.parse import urlparse, parse_qs, parse_qsl
    # pylint: enable=E0611,F0401
except ImportError:
    from urlparse import urlparse, parse_qs, parse_qsl

from twisted.internet import defer
from ubuntuone.devtools.handlers import MementoHandler
from ubuntuone.devtools.testcases import TestCase

from ubuntu_kylin_sso.utils.webclient import restful
from ubuntu_kylin_sso.utils.webclient.common import Response


SAMPLE_SERVICE_IRI = u"http://localhost/"
SAMPLE_NAMESPACE = u"sample_namespace"
SAMPLE_METHOD = u"sample_method"
SAMPLE_OPERATION = SAMPLE_NAMESPACE + u"." + SAMPLE_METHOD
SAMPLE_ARGS = dict(uno=1, dos=u"2", tres=u"ñandú")
SAMPLE_RESPONSE = restful.json.dumps(SAMPLE_ARGS)
SAMPLE_USERNAME = "joeuser@example.com"
SAMPLE_PASSWORD = "clavesecreta"
SAMPLE_OAUTH_CREDS = dict(token="1234", etc="456")


class FakeWebClient(object):
    """A fake web client."""

    def __init__(self, **kwargs):
        """Initialize this faker."""
        self.return_value = SAMPLE_RESPONSE
        self.called = []
        self.init_kwargs = kwargs
        self.running = True

    def request(self, iri, *args, **kwargs):
        """Return a deferred that will be fired with a Response object."""
        self.called.append((iri, args, kwargs))
        return defer.succeed(Response(self.return_value))

    def shutdown(self):
        """Stop this fake webclient."""
        self.running = False


class BaseTestCase(TestCase):
    """The base for the Restful Client testcases."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this test case."""
        yield super(BaseTestCase, self).setUp()
        self.wc = None
        self.patch(restful.webclient, "webclient_factory",
                   self.webclient_factory)

    def webclient_factory(self, **kwargs):
        """A factory that saves the webclient created."""
        self.wc = FakeWebClient(**kwargs)
        return self.wc


class RestfulClientTestCase(BaseTestCase):
    """Tests for the proxy-enabled Restful Client."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this testcase."""
        yield super(RestfulClientTestCase, self).setUp()
        self.rc = restful.RestfulClient(SAMPLE_SERVICE_IRI)
        self.addCleanup(self.rc.shutdown)

    def test_has_a_webclient(self):
        """The RC has a webclient."""
        self.assertEqual(self.rc.webclient, self.wc)

    def test_shutsdown_the_webclient(self):
        """Calling shutdown on the restful shuts down the webclient too."""
        self.rc.shutdown()
        # pylint: disable=E1101
        self.assertFalse(self.rc.webclient.running, "The webclient is stopped")

    @defer.inlineCallbacks
    def test_can_make_calls(self):
        """The RC can make webcalls."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        self.assertEqual(len(self.wc.called), 1)

    @defer.inlineCallbacks
    def test_restful_namespace_added_to_url(self):
        """The restful namespace is added to the url."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        iri, _, _ = self.wc.called[0]
        uri = iri.encode("ascii")
        url = urlparse(uri)
        # pylint: disable=E1101,E1103
        self.assertTrue(url.path.endswith(SAMPLE_NAMESPACE),
                        "The namespace is included in url")

    @defer.inlineCallbacks
    def test_restful_method_added_to_params(self):
        """The restful method is added to the params."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        _, _, webcall_kwargs = self.wc.called[0]
        wc_params = parse_qs(webcall_kwargs["post_content"])
        self.assertEqual(wc_params["ws.op"][0], SAMPLE_METHOD)

    @defer.inlineCallbacks
    def test_arguments_added_as_json_to_webcall(self):
        """The keyword arguments are used as json in the webcall."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        _, _, webcall_kwargs = self.wc.called[0]
        params = parse_qsl(webcall_kwargs["post_content"])
        result = {}
        for key, value in params:
            if key == "ws.op":
                continue
            result[key] = restful.json.loads(value)
        self.assertEqual(result, SAMPLE_ARGS)

    @defer.inlineCallbacks
    def test_post_header_sent(self):
        """A header is sent specifying the contents of the post."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        _, _, webcall_kwargs = self.wc.called[0]
        self.assertEqual(restful.POST_HEADERS,
                         webcall_kwargs["extra_headers"])

    @defer.inlineCallbacks
    def test_post_method_set(self):
        """The method of the webcall is set to POST."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        _, _, webcall_kwargs = self.wc.called[0]
        self.assertEqual("POST", webcall_kwargs["method"])

    @defer.inlineCallbacks
    def test_return_value_json_parsed(self):
        """The result is json parsed before being returned."""
        result = yield self.rc.restcall(SAMPLE_OPERATION)
        self.assertEqual(result, SAMPLE_ARGS)


class AuthenticationOptionsTestCase(BaseTestCase):
    """Tests for the authentication options."""

    def test_passes_userpass_to_webclient_init(self):
        """The RestfulClient passes the user and pass to the webclient."""
        params = dict(username=SAMPLE_USERNAME, password=SAMPLE_PASSWORD)
        restful.RestfulClient(SAMPLE_SERVICE_IRI, **params)
        expected = dict(params)
        expected["oauth_sign_plain"] = True
        self.assertEqual(self.wc.init_kwargs, expected)

    @defer.inlineCallbacks
    def test_passes_oauth_creds_to_request(self):
        """The RestfulClient passes the credentials in each request."""
        kwargs = dict(oauth_credentials=SAMPLE_OAUTH_CREDS)
        rc = restful.RestfulClient(SAMPLE_SERVICE_IRI, **kwargs)
        yield rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)
        _, _, kwargs = self.wc.called[0]
        self.assertEqual(kwargs["oauth_credentials"], SAMPLE_OAUTH_CREDS)


class LogginTestCase(BaseTestCase):
    """Ensure that proper debug logging is done."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this testcase."""
        yield super(LogginTestCase, self).setUp()
        self.memento = MementoHandler()
        restful.logger.addHandler(self.memento)
        restful.logger.setLevel(logging.DEBUG)
        self.addCleanup(restful.logger.removeHandler, self.memento)

        self.rc = restful.RestfulClient(SAMPLE_SERVICE_IRI)
        self.addCleanup(self.rc.shutdown)

    @defer.inlineCallbacks
    def test_log_rest_call(self):
        """Check that proper DEBUG is made for every REST call."""
        yield self.rc.restcall(SAMPLE_OPERATION, **SAMPLE_ARGS)

        expected_msgs = (
            SAMPLE_SERVICE_IRI + SAMPLE_NAMESPACE,
        )
        self.assertTrue(self.memento.check_debug(*expected_msgs))

    @defer.inlineCallbacks
    def test_log_json_loads_exception(self):
        """Check that json load errors are properly logged."""
        invalid_json = 'NOTAVALIDJSON'
        self.patch(self.wc, 'return_value', invalid_json)
        yield self.assertFailure(self.rc.restcall(SAMPLE_OPERATION),
                                 ValueError)

        self.memento.debug = True
        expected_msgs = (
            ValueError,
            'Can not load json from REST request response',
            invalid_json
        )
        self.assertTrue(self.memento.check_exception(*expected_msgs))



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""A Expander widget similar to the GtkExpander."""

from PyQt4.QtCore import pyqtSignal
from PyQt4.QtGui import QHBoxLayout, QLabel, QSizePolicy, QVBoxLayout, QWidget

from ubuntu_kylin_sso.qt.arrow import QArrow

# we are following the Qt style, lets tell pylint to ignore it
# pylint: disable=C0103


class QExpanderLabel(QWidget):
    """Widget used to show the label of a QExpander."""

    clicked = pyqtSignal()

    def __init__(self, label, parent=None):
        """Create a new instance."""
        super(QExpanderLabel, self).__init__(parent)
        self.arrow = QArrow(QArrow.RIGHT)
        self.label = QLabel(label)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        layout.addWidget(self.arrow)
        layout.addWidget(self.label)

    def mousePressEvent(self, event):
        """Mouse clicked."""
        if self.arrow.direction == QArrow.DOWN:
            self.arrow.direction = QArrow.RIGHT
        else:
            self.arrow.direction = QArrow.DOWN
        self.clicked.emit()

    def text(self):
        """Return the text of the label."""
        return self.label.text()

    def setText(self, text):
        """Set the text of the label."""
        self.label.setText(text)


class QExpander(QWidget):
    """A Qt implementation similar to GtkExpander."""

    def __init__(self, label, expanded=False, parent=None):
        """Create a new instance."""
        super(QExpander, self).__init__(parent)
        self.label = QExpanderLabel(label)
        self.label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.content = None
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.layout.addWidget(self.label)
        self.layout.addStretch()
        self.label.clicked.connect(self._on_label_clicked)
        self.setExpanded(expanded)

    def _on_label_clicked(self):
        """The expander widget was clicked."""
        self._expanded = not self._expanded
        self.setExpanded(self._expanded)

    def addWidget(self, widget):
        """Add a widget to the expander.

        The previous widget will be removed.
        """
        if self.content is not None:
            self.layout.removeWidget(self.content)
        self.content = widget
        self.content.setVisible(self._expanded)
        self.layout.insertWidget(1, self.content)

    def text(self):
        """Return the text of the label."""
        return self.label.text()

    def setText(self, text):
        """Set the text of the label."""
        self.label.setText(text)

    def expanded(self):
        """Return if widget is expanded."""
        return self._expanded

    # pylint: disable=W0201
    def setExpanded(self, is_expanded):
        """Expand the widget or not."""
        self._expanded = is_expanded
        if self._expanded:
            self.label.arrow.direction = QArrow.DOWN
        else:
            self.label.arrow.direction = QArrow.RIGHT
        if self.content is not None:
            self.content.setVisible(self._expanded)
    # pylint: enable=W0201



# -*- coding: utf-8 -*-
#
# ubuntu_kylin_sso.logger - logging miscellany
#
# Author: Stuart Langridge <stuart.langridge@canonical.com>
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright 2009-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Miscellaneous logging functions."""

import logging
import os
import sys

from dirspec.basedir import xdg_cache_home
from dirspec.utils import unicode_path
from functools import wraps
from logging.handlers import RotatingFileHandler

# LOGFOLDER = os.path.join(xdg_cache_home, 'sso')
LOGFOLDER = os.path.join(xdg_cache_home, 'ubuntu_kylin_sso')
# create log folder if it doesn't exists
if not os.path.exists(unicode_path(LOGFOLDER)):
    os.makedirs(unicode_path(LOGFOLDER))

if os.environ.get('U1_DEBUG'):
    LOG_LEVEL = logging.DEBUG
else:
    # Only log this level and above
    LOG_LEVEL = logging.INFO

LOG_PATH = os.path.join(LOGFOLDER, 'sso-client.log')
FMT = "%(asctime)s:%(msecs)s - %(name)s - %(levelname)s - %(message)s"

MAIN_HANDLER = RotatingFileHandler(unicode_path(LOG_PATH),
                                   maxBytes=1048576,
                                   backupCount=5)
MAIN_HANDLER.setLevel(LOG_LEVEL)
MAIN_HANDLER.setFormatter(logging.Formatter(fmt=FMT))

GUI_LOG_PATH = os.path.join(LOGFOLDER, 'sso-client-gui.log')
GUI_HANDLER = RotatingFileHandler(unicode_path(GUI_LOG_PATH),
                                  maxBytes=1048576,
                                  backupCount=5)
GUI_HANDLER.setLevel(LOG_LEVEL)
GUI_HANDLER.setFormatter(logging.Formatter(fmt=FMT))


def setup_logging(log_domain, handler=None):
    """Create basic logger to set filename."""
    if handler is None:
        handler = MAIN_HANDLER

    logger = logging.getLogger(log_domain)
    logger.propagate = False
    logger.setLevel(LOG_LEVEL)
    logger.addHandler(handler)
    if os.environ.get('U1_DEBUG'):
        debug_handler = logging.StreamHandler(sys.stderr)
        debug_handler.setFormatter(logging.Formatter(fmt=FMT))
        logger.addHandler(debug_handler)

    return logger


def setup_gui_logging(log_domain):
    """Create basic logger to set filename."""
    return setup_logging(log_domain, GUI_HANDLER)


def log_call(log_func):
    """Decorator to log, using 'log_func', calls to functions."""

    def middle(f):
        """Return a function that will act as 'f' but will log the call."""

        @wraps(f)
        def inner(instance, *a, **kw):
            """Call 'f(*a, **kw)' and return its result. Log that call."""
            log_func('%r: emitting %r with args %r and kwargs %r',
                     instance.__class__.__name__, f.__name__, a, kw)
            result = f(instance, *a, **kw)
            return result

        return inner

    return middle



# -*- coding: utf-8 -*-
#
# test_linux - tests for ubuntu_kylin_sso.keyring under linux
#
# Author: Alejandro J. Cura <alecu@canonical.com>
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the keyring.py module."""

from __future__ import unicode_literals

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso import keyring as common_keyring
from ubuntu_kylin_sso.keyring import linux as keyring


class MockItem(object):
    """An item contains a secret, lookup attributes and has a label."""

    def __init__(self, label, collection, attr, value):
        """Initialize a new Item."""
        self.label = label
        self.collection = collection
        self.attributes = attr
        self.value = value

    def get_value(self):
        """Retrieve the secret for this item."""
        return defer.succeed(self.value)

    def delete(self):
        """Delete this item."""
        self.collection.items.remove(self)
        return defer.succeed(None)

    def matches(self, search_attr):
        """See if this item matches a given search."""
        for k, val in search_attr.items():
            if k not in self.attributes:
                return False
            if self.attributes[k] != val:
                return False
        return True


class MockCollection(object):
    """A collection of items containing secrets."""

    def __init__(self, label, service):
        """Initialize a new collection."""
        self.label = label
        self.service = service
        self.items = []

    def create_item(self, label, attr, value, replace=True):
        """Create an item with the given attributes, secret and label."""
        item = MockItem(label, self, attr, value)
        self.items.append(item)
        return defer.succeed(item)


class MockSecretService(object):
    """A class that mocks txsecrets.SecretService."""

    def __init__(self, *args, **kwargs):
        super(MockSecretService, self).__init__(*args, **kwargs)
        self.collections = {}

    def open_session(self, window_id=0):
        """Open a unique session for the caller application."""
        return defer.succeed(self)

    def search_items(self, attributes):
        """Find items in any collection."""
        results = []
        for collection in self.collections.values():
            for item in collection.items:
                if item.matches(attributes):
                    results.append(item)
        return defer.succeed(results)

    def create_collection(self, label):
        """Create a new collection with the specified properties."""
        collection = MockCollection(label, self)
        self.collections[label] = collection
        if "default" not in self.collections:
            self.collections["default"] = collection
        return defer.succeed(collection)

    def get_default_collection(self):
        """The collection were default items should be created."""
        if len(self.collections) == 0:
            self.create_collection("default")
        return defer.succeed(self.collections["default"])


class TestKeyring(TestCase):
    """Test the keyring related functions."""

    timeout = 5

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize the mock used in these tests."""
        yield super(TestKeyring, self).setUp()
        self.mock_service = None
        self.service = self.patch(keyring, "SecretService",
                                  self.get_mock_service)
        self.patch(common_keyring, "gethostname", lambda: "darkstar")

    def get_mock_service(self):
        """Create only one instance of the mock service per test."""
        if self.mock_service is None:
            self.mock_service = MockSecretService()
        return self.mock_service

    @inlineCallbacks
    def test_set_credentials(self):
        """Test that the set method does not erase previous keys."""
        sample_creds = {"name": "sample creds name"}
        sample_creds2 = {"name": "sample creds name 2"}
        kr = keyring.Keyring()
        yield kr.set_credentials("appname", sample_creds)
        yield kr.set_credentials("appname", sample_creds2)

        # pylint: disable=E1101
        self.assertEqual(len(kr.service.collections["default"].items), 2)

    @inlineCallbacks
    def test_delete_credentials(self):
        """Test that a given key is deleted."""
        sample_creds = {"name": "sample creds name"}
        kr = keyring.Keyring()
        yield kr.set_credentials("appname", sample_creds)
        yield kr.delete_credentials("appname")

        # pylint: disable=E1101
        self.assertEqual(len(kr.service.collections["default"].items), 1)

    @inlineCallbacks
    def test_get_credentials(self):
        """Test that credentials are properly retrieved."""
        sample_creds = {"name": "sample creds name"}
        kr = keyring.Keyring()
        yield kr.set_credentials("appname", sample_creds)

        result = yield kr.get_credentials("appname")
        self.assertEqual(result, sample_creds)

    @inlineCallbacks
    def test_get_credentials_migrating_token(self):
        """Test that credentials are properly retrieved and migrated."""
        sample_creds = {"name": "sample creds name"}
        kr = keyring.Keyring()
        self.patch(keyring, "get_token_name", keyring.get_old_token_name)
        yield kr.set_credentials("app name", sample_creds)

        result = yield kr.get_credentials("app name")
        self.assertEqual(result, sample_creds)

    @inlineCallbacks
    def test_get_old_cred_found(self):
        """The method returns a new set of creds if old creds are found."""
        sample_oauth_token = "sample oauth token"
        sample_oauth_secret = "sample oauth secret"
        old_creds = {
            "oauth_token": sample_oauth_token,
            "oauth_token_secret": sample_oauth_secret,
        }
        u1kr = common_keyring.UbuntuOneOAuthKeyring()
        yield u1kr.set_credentials(keyring.U1_APP_NAME, old_creds)

        kr = keyring.Keyring()
        result = yield kr.get_credentials(keyring.U1_APP_NAME)
        self.assertIn("token", result)
        self.assertEqual(result["token"], sample_oauth_token)
        self.assertIn("token_secret", result)
        self.assertEqual(result["token_secret"], sample_oauth_secret)

    @inlineCallbacks
    def test_get_old_cred_found_but_not_asked_for(self):
        """Returns None if old creds are present but the appname is not U1"""
        sample_oauth_token = "sample oauth token"
        sample_oauth_secret = "sample oauth secret"
        old_creds = {
            "oauth_token": sample_oauth_token,
            "oauth_token_secret": sample_oauth_secret,
        }
        u1kr = common_keyring.UbuntuOneOAuthKeyring()
        yield u1kr.set_credentials(keyring.U1_APP_NAME, old_creds)

        kr = keyring.Keyring()
        result = yield kr.get_credentials("Software Center")
        self.assertEqual(result, None)

    @inlineCallbacks
    def test_get_old_cred_not_found(self):
        """The method returns None if no old nor new credentials found."""
        kr = keyring.Keyring()
        result = yield kr.get_credentials(keyring.U1_APP_NAME)
        self.assertEqual(result, None)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Qt implementation of proxy UI."""

import argparse
import sys

from PyQt4.QtGui import QApplication, QDialog, QIcon
from twisted.internet import defer

from ubuntu_kylin_sso import EXCEPTION_RAISED, USER_SUCCESS, USER_CANCELLATION
from ubuntu_kylin_sso.logger import setup_gui_logging
from ubuntu_kylin_sso.keyring import Keyring
from ubuntu_kylin_sso.qt.ui.proxy_credentials_dialog_ui import Ui_ProxyCredsDialog
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    CANCEL_BUTTON,
    PROXY_CREDS_DIALOG_TITLE,
    PROXY_CREDS_HEADER,
    PROXY_CREDS_EXPLANATION,
    PROXY_CREDS_CONNECTION,
    PROXY_CREDS_ERROR,
    PROXY_CREDS_USER_LABEL,
    PROXY_CREDS_PSWD_LABEL,
    PROXY_CREDS_HELP_BUTTON,
    PROXY_CREDS_SAVE_BUTTON,
)

logger = setup_gui_logging("ubuntu_kylin_sso.qt.proxy_dialog")


class ProxyCredsDialog(QDialog):
    """Dialog used to require the proxy credentials."""

    def __init__(self, retry=False, domain=None):
        """Create a new instance."""
        super(ProxyCredsDialog, self).__init__()

        if domain is None:
            logger.debug('Domain passed as None.')
            domain = ''
        self.domain = domain
        self.keyring = Keyring()
        self.ui = Ui_ProxyCredsDialog()
        self.ui.setupUi(self)
        # lets set the different basic contents for the ui
        self._set_labels()
        self._set_buttons()
        self._set_icon()
        if retry:
            self._load_creds()
            self.ui.error_label.setVisible(True)
        else:
            self.ui.error_label.setVisible(False)

    @defer.inlineCallbacks
    def _load_creds(self):
        """Tries to load the creds in a retry event."""
        # pylint: disable=W0703
        try:
            creds = yield self.keyring.get_credentials(self.domain)
            if creds is not None:
                logger.debug('Go no empty credentials.')
                # lets set the text for the inputs
                self.ui.username_entry.setText(creds['username'])
                self.ui.password_entry.setText(creds['password'])
        except Exception:
            logger.error('Problem getting old creds.')
        # pylint: enable=W0703

    def _set_labels(self):
        """Set the labels translations."""
        self.setWindowTitle(PROXY_CREDS_DIALOG_TITLE)
        self.ui.title_label.setText(PROXY_CREDS_HEADER)
        self.ui.explanation_label.setText(PROXY_CREDS_EXPLANATION)
        self.ui.connection_label.setText(PROXY_CREDS_CONNECTION)
        # HACK: later this should be set using qss
        self.ui.error_label.setText("<font color='#a62626'><b>%s</b></font>"
                                     % PROXY_CREDS_ERROR)
        self.ui.username_label.setText(PROXY_CREDS_USER_LABEL)
        self.ui.password_label.setText(PROXY_CREDS_PSWD_LABEL)
        self.ui.domain_label.setText(self.domain)

    @defer.inlineCallbacks
    def _on_save_clicked(self, *args):
        """Save the new credentials."""
        username = compat.text_type(
                        self.ui.username_entry.text()).encode('utf8')
        password = compat.text_type(
                        self.ui.password_entry.text()).encode('utf8')
        creds = dict(username=username, password=password)
        try:
            logger.debug('Save credentials as for domain %s.', self.domain)
            yield self.keyring.set_credentials(self.domain, creds)
        except Exception as e:
            logger.exception('Could not set credentials: %s', e)
            self.done(EXCEPTION_RAISED)
        logger.debug('Stored creds')
        self.done(USER_SUCCESS)

    def _on_cancel_clicked(self, *args):
        """End the dialog."""
        logger.debug('User canceled credentials dialog.')
        self.done(USER_CANCELLATION)

    def _set_buttons(self):
        """Set the labels of the buttons."""
        self.ui.help_button.setText(PROXY_CREDS_HELP_BUTTON)
        self.ui.cancel_button.setText(CANCEL_BUTTON)
        self.ui.cancel_button.clicked.connect(self._on_cancel_clicked)
        self.ui.save_button.setText(PROXY_CREDS_SAVE_BUTTON)
        self.ui.save_button.clicked.connect(self._on_save_clicked)

    def _set_icon(self):
        """Set the icon used in the dialog."""
        icon = QIcon.fromTheme('gtk-dialog-authentication')
        self.ui.logo_label.setText('')
        self.ui.logo_label.setPixmap(icon.pixmap(48, 48))


def parse_args():
    """Parse sys.arg options."""
    parser = argparse.ArgumentParser(
            description='Open the Qt Proxy Credentials UI.')
    parser.add_argument('--domain', required=True,
        help='the domain whose credentials are going to be stored.')
    parser.add_argument('--retry', action='store_true', default=False,
        help='whether we are retrying to get the creds.')
    return parser.parse_args()


def exit_code(return_code):
    """Use the window result code and the sys.exit."""
    logger.debug('exit %s', return_code)
    QApplication.instance().exit(return_code)
    if sys.platform == 'win32':
        logger.debug('Stop qt reactor')
        from twisted.internet import reactor
        reactor.stop()


def main():
    """Main method used to show the creds dialog."""
    if sys.platform == 'win32':
        import qt4reactor
        qt4reactor.install()
        logger.debug('Qt reactor installed.')

    app = QApplication(sys.argv)
    app
    args = parse_args()
    win = ProxyCredsDialog(domain=args.domain,
                           retry=args.retry)

    if sys.platform == 'win32':
        win.show()
        win.finished.connect(exit_code)

        logger.debug('Starting reactor')
        from twisted.internet import reactor
        logger.debug('QApp is %s', reactor.qApp)
        reactor.run()
    else:
        return_code = win.exec_()
        sys.exit(return_code)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
A test script to start a NetworkManagerState instance and print out
the current actual state of the system.
"""

from __future__ import print_function

from ubuntu_kylin_sso.networkstate import (NetworkManagerState,
                                     is_machine_connected)


def my_cb(state):
    """simple callback just prints state"""
    print("login.ubuntu.com:", state)

if __name__ == '__main__':
    # test once with direct call:
    is_machine_connected().addCallback(my_cb)

    NMS = NetworkManagerState(my_cb)

    NMS.find_online_state()

    import time
    time.sleep(60)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Windows specific tests for the main module."""

from twisted.internet import defer

from ubuntu_kylin_sso.main import windows
from ubuntu_kylin_sso.tests import TestCase

# because we are using twisted we have java like names C0103
# pylint: disable=C0103


class MockWin32APIs(object):
    """Some mock win32apis."""

    process_handle = object()
    TOKEN_ALL_ACCESS = object()
    TokenUser = object()

    def __init__(self, sample_token):
        """Initialize this mock instance."""
        self.sample_token = sample_token
        self.token_handle = object()

    def GetCurrentProcess(self):
        """Returns a fake process_handle."""
        return self.process_handle

    def OpenProcessToken(self, process_handle, access):
        """Open the process token."""
        assert process_handle is self.process_handle
        assert access is self.TOKEN_ALL_ACCESS
        return self.token_handle

    def GetTokenInformation(self, token_handle, info):
        """Get the information for this token."""
        assert token_handle == self.token_handle
        assert info == self.TokenUser
        return (self.sample_token, 0)

# pylint: enable=C0103


class MiscTestCase(TestCase):
    """Tests for module level misc functions."""

    def test_get_userid(self):
        """The returned user id is parsed ok."""
        expected_id = 1001
        sample_token = "abc-123-1001"

        win32apis = MockWin32APIs(sample_token)
        self.patch(windows, "win32process", win32apis)
        self.patch(windows, "win32security", win32apis)

        userid = windows.get_user_id()
        self.assertEqual(userid, expected_id)

    def _test_port_assignation(self, uid, expected_port):
        """Test a given uid/expected port combo."""
        self.patch(windows, "get_user_id", lambda: uid)
        self.assertEqual(windows.get_sso_pb_port(), expected_port)

    def test_get_sso_pb_port(self):
        """Test the get_sso_pb_port function."""
        uid = 1001
        uid_modulo = uid % windows.SSO_RESERVED_PORTS
        expected_port = (windows.SSO_BASE_PB_PORT +
                         uid_modulo * windows.SSO_PORT_ALLOCATION_STEP)
        self._test_port_assignation(uid, expected_port)

    def test_get_sso_pb_port_alt(self):
        """Test the get_sso_pb_port function."""
        uid = 2011 + windows.SSO_RESERVED_PORTS
        uid_modulo = uid % windows.SSO_RESERVED_PORTS
        expected_port = (windows.SSO_BASE_PB_PORT +
                         uid_modulo * windows.SSO_PORT_ALLOCATION_STEP)
        self._test_port_assignation(uid, expected_port)


class DescriptionFactoryTestcase(TestCase):
    """Test the factory."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the tests."""
        yield super(DescriptionFactoryTestcase, self).setUp()
        self.port = 55555
        self.patch(windows, 'get_sso_pb_port', lambda: self.port)

    def test_server_description(self):
        """Test getting the description info."""
        expected = windows.DescriptionFactory.server_description_pattern % \
                       self.port
        factory = windows.DescriptionFactory()
        self.assertEqual(expected, factory.server)

    def test_client_description(self):
        """Test getting the description info."""
        expected = windows.DescriptionFactory.client_description_pattern % \
                       self.port
        factory = windows.DescriptionFactory()
        self.assertEqual(expected, factory.client)



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""The Qt graphical interface for the Ubuntu Single Sign On Client."""

from __future__ import unicode_literals

import sys
import collections

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.logger import setup_gui_logging
from ubuntu_kylin_sso.utils.ui import GENERIC_BACKEND_ERROR, WINDOW_TITLE_COMMON


logger = setup_gui_logging('ubuntu_kylin_sso.qt')

LINK_STYLE = ('<a href="{link_url}">'
              '<span style="color:#df2d1f;">{link_text}</span></a>')
ERROR_ALL = '__all__'
ERROR_STYLE = '<font color="#df2d1f" style="font-size:small"><b>%s</b></font>'
ERROR_MESSAGE = 'message'
PREFERED_UI_SIZE = {'width': 550, 'height': 525}
TITLE_STYLE = '<span style="font-size:xx-large;font-weight:bold;">%s</span>'
#WINDOW_TITLE = 'Ubuntu Kylin Single Sign On'
WINDOW_TITLE = WINDOW_TITLE_COMMON

# TODO: There is a pixel discrepancy between Qt on Linux and Windows
# and Mac OS X. On Mac OS X, one test fails with the height being
# off. For now, we're forcing a different UI height on darwin.
if sys.platform == 'darwin':
    PREFERED_UI_SIZE['height'] = 533


def build_general_error_message(errordict):
    """Build a user-friendly error message from the errordict."""
    logger.debug('build_general_error_message: errordict is: %r.', errordict)
    result = ''
    if isinstance(errordict, collections.Mapping):
        msg1 = errordict.get(ERROR_ALL)
        msg2 = errordict.get(ERROR_MESSAGE)
        if msg2 is None:
            # See the errordict in LP: 828417
            msg2 = errordict.get('error_message')
        if msg1 is not None and msg2 is not None:
            result = '\n'.join((msg1, msg2))
        elif msg1 is not None:
            result = msg1
        elif msg2 is not None:
            result = msg2
        else:
            if 'errtype' in errordict:
                del errordict['errtype']
            result = '\n'.join(
                [('%s: %s' % (k, v)) for k, v in errordict.items()])
    else:
        result = GENERIC_BACKEND_ERROR
        logger.error('build_general_error_message with unknown error: %r',
            errordict)

    logger.info('build_general_error_message: returning %r.', result)
    return result


def maybe_elide_text(label, text, width, markup=None):
    """Set 'text' to be the 'label's text.

    If 'text' is longer than 'width', set the label's tooltip to be the full
    text, and the text itself to be the elided version of 'text'.

    """
    fm = QtGui.QFontMetrics(label.font())
    elided_text = fm.elidedText(text, QtCore.Qt.ElideRight, width)
    if elided_text != text:
        label.setToolTip(text)
    if markup is not None:
        elided_text = markup % elided_text
    label.setText(elided_text)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Forgotten Password page UI."""

from functools import partial

from PyQt4 import QtCore

from ubuntu_kylin_sso import NO_OP
from ubuntu_kylin_sso.logger import setup_gui_logging, log_call
from ubuntu_kylin_sso.qt.sso_wizard_page import SSOWizardEnhancedEditPage
from ubuntu_kylin_sso.qt.ui.forgotten_password_ui import Ui_ForgottenPasswordPage
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    EMAIL_LABEL,
    FORGOTTEN_PASSWORD_TITLE,
    FORGOTTEN_PASSWORD_SUBTITLE,
    is_correct_email,
    RESET_PASSWORD,
    REQUEST_PASSWORD_TOKEN_WRONG_EMAIL,
)


logger = setup_gui_logging('ubuntu_kylin_sso.forgotten_password_page')


class ForgottenPasswordPage(SSOWizardEnhancedEditPage):
    """Widget used to deal with users that forgot the password."""

    ui_class = Ui_ForgottenPasswordPage
    passwordResetTokenSent = QtCore.pyqtSignal(compat.text_type)

    @property
    def _signals(self):
        """The signals to connect to the backend."""
        result = {
            'PasswordResetTokenSent':
                self._filter_by_app_name(self.on_password_reset_token_sent),
            'PasswordResetError':
                self._filter_by_app_name(self.on_password_reset_error),
        }
        return result

    @property
    def email_address(self):
        """Return the email address provided by the user."""
        return compat.text_type(self.ui.email_line_edit.text())

    # pylint: disable=C0103

    def initializePage(self):
        """Set the initial state of ForgottenPassword page."""
        logger.debug('initializePage - About to show ForgottenPasswordPage')
        self.ui.send_button.setDefault(True)
        enabled = not self.ui.email_line_edit.text().isEmpty()
        self.ui.send_button.setEnabled(enabled)

    # pylint: enable=C0103

    def _register_fields(self):
        """Register the fields of the wizard page."""
        self.registerField('email_address',
                           self.ui.email_line_edit)

    def _set_translated_strings(self):
        """Set the translated strings in the view."""
        self.setTitle(FORGOTTEN_PASSWORD_TITLE)
        subtitle = FORGOTTEN_PASSWORD_SUBTITLE.format(app_name=self.app_name)
        self.setSubTitle(subtitle)
        self.ui.email_address_label.setText(EMAIL_LABEL)
        self.ui.send_button.setText(RESET_PASSWORD)

    def _set_enhanced_line_edit(self):
        """Set the extra logic to the line edits."""
        self.set_line_edit_validation_rule(self.ui.email_line_edit,
                                           is_correct_email)

    def _connect_ui(self):
        """Connect the diff signals from the Ui."""
        self.ui.email_line_edit.textChanged.connect(self._validate)
        self.ui.send_button.clicked.connect(self.request_new_password)
        self._set_enhanced_line_edit()
        self._register_fields()

    def request_new_password(self):
        """Send the request password operation."""
        self.hide_error()
        args = (self.app_name, self.email_address)
        logger.debug('Sending request new password for %s, email: %s', *args)
        f = self.backend.request_password_reset_token

        error_handler = partial(self._handle_error, f,
            self.on_password_reset_error)

        self.show_overlay()
        f(*args, reply_handler=NO_OP, error_handler=error_handler)

    def _validate(self):
        """Validate that we have an email."""
        email = compat.text_type(self.ui.email_line_edit.text())
        self.ui.send_button.setEnabled(is_correct_email(email))

    def on_password_reset_token_sent(self, app_name, email):
        """Action taken when we managed to get the password reset done."""
        logger.info('ForgottenPasswordPage.on_password_reset_token_sent for '
                    '%s, email: %s', app_name, email)
        # ignore the result and move to the reset page
        self.hide_overlay()
        self.passwordResetTokenSent.emit(email)

    @log_call(logger.error)
    def on_password_reset_error(self, app_name, error):
        """Action taken when there was an error requesting the reset."""
        # set the error message
        self.hide_overlay()
        msg = REQUEST_PASSWORD_TOKEN_WRONG_EMAIL
        self.show_error(msg)
